/* Copyright (C) David Wolfe, 1990.  All rights reserved. */

#include <stdlib.h>
#include <stdio.h>
#include "externs.h"
#include "list.h"
#include "bitmatrx.h"

int	count_B_frees = 0;
int	count_B_makes = 0;
int	symmetries_for_4[16];

#define DO_REFLECT_X_AXIS	((REFLECT_X_AXIS << 3) | REFLECT_X_AXIS)
#define DO_REFLECT_Y_AXIS	((REFLECT_Y_AXIS << 3) | REFLECT_Y_AXIS)
#define DO_TRANSPOSE      	((TRANSPOSE << 3) | TRANSPOSE)
#define DO_NOT_REFLECT_X_AXIS	(REFLECT_X_AXIS << 3)
#define DO_NOT_REFLECT_Y_AXIS	(REFLECT_Y_AXIS << 3)
#define DO_NOT_TRANSPOSE      	(TRANSPOSE << 3)

B_type B_make (int x, int y)
{
	B_type a;
	int i;

	count_B_makes++;
	if (!(a = (B_type) malloc (sizeof (struct B_struct))))
		myerror ("B_make: malloc couldn't free block");
	a->x = x;
	a->y = y;
	if (!(a->contents = (unsigned *) malloc (((x*y+(INT_SIZE - 1)) / INT_SIZE) * sizeof (unsigned))))
		myerror ("B_make: malloc couldn't free block");
	for (i = 0; i < (x*y+(INT_SIZE - 1))/INT_SIZE; i++) *(a->contents + i) = (unsigned) 0;
	return a;
}

list_type B_to_list (B_type a)
{
	list_type l;
	int i;
	
	l = list_make();
	for (i = 0; i < (a->x * a->y+(INT_SIZE-1)) / INT_SIZE; i++)
	    list_prepend (l, *(a->contents + i));
	list_prepend (l, (a->x << 8 | a->y));
	return l;
}

B_type B_pack (B_type a)
{
	B_type b;
	int i, j, xmin, ymin, xmax, ymax;

	xmin=a->x, ymin=a->y, xmax=-1, ymax=-1;
	for (i=0; i < a->x; i++)
		for (j=0; j < a->y; j++)
			if (B_get (a,i,j)) {
				if (i < xmin) xmin = i;
				if (j < ymin) ymin = j;
				if (i > xmax) xmax = i;
				if (j > ymax) ymax = j;
			}
	if (xmin == 0 && ymin==0 && xmax==a->x-1 && ymax==a->y-1) return a;
	if (xmin > xmax) {
		B_free (a);
		return B_make (0,0);
	}
	b = B_make (xmax - xmin + 1, ymax - ymin + 1);
	for (i = 0; i < xmax - xmin + 1; i++)
		for (j = 0; j < ymax - ymin + 1; j++)
			if (B_get (a, i+xmin, j+ymin)) B_set (b,i,j);
	B_free (a);
	return b;
}

void B_printf (B_type a) {
    B_printf_with_symbols (a, "0", "1");
}

void B_printf_with_symbols (B_type a, char *zero, char *one) {
	int x,y;
	for (y = 0; y < a->y; y++) {
		for (x = 0; x < a->x; x++)
			printf ("%s", B_get (a,x,y) ? one : zero);
		printf ("\n");
	}
}

B_type B_copy (B_type a)
{
  B_type b;
  int i,n, x = B_x(a), y=B_y(a);
  

  b = B_make(x,y);
  n = (x*y == 0) ? -1 : (x*y - 1) / INT_SIZE;
  for (i=0; i <= n; i++)
    *(b->contents + i) = *(a->contents + i);
  return b;
}

B_type B_copy_into (B_type b, B_type a)
{
	int i, j;
	
	for (i=0; i<B_x(a); i++)
		for (j=0; j<B_y(a); j++)
			if (B_get (a,i,j))
				B_set (b,i,j);
			else
				B_clear (b,i,j);
	return b;
}

B_type B_copy_offset (B_type old, int xoffset, int yoffset, int xsize, int ysize)
{
  B_type New;
  int i, j, k, l;

  New = B_make(xsize, ysize);

  for (i=0, k=xoffset; i<xsize; i++, k++)
    for (j=0, l=yoffset; j<ysize; j++, l++)
      if (B_get(old, k, l))
	B_set(New, i, j);
  return New;
}


/* Does not modify or free its argument; just returns an expanded copy. */
B_type B_expand (B_type a, int dx, int dy)
{
  B_type b;
  int i, j;

  b = B_make (B_x(a) + abs(dx), B_y(a) + abs(dy));
  if (dx + dy > 0)
    { for (i=0; i < B_x(a); i++)
      for (j=0; j < B_y(a); j++)
	if (B_get (a,i,j))
	  B_set (b,i,j); }
  else
    /* dx or dy is -1, the other is 0 */
    { for (i=0; i < B_x(a); i++)
      for (j=0; j < B_y(a); j++)
	  if (B_get (a,i,j))
	    B_set (b,i-dx,j-dy); }
  return b;
}

B_type B_transpose (B_type a)
{
  B_type b;
  int i, j;

  b = B_make(B_y(a), B_x(a));

  for (i=0; i<B_x(a); i++)
    for (j=0; j<B_y(a); j++)
      if (B_get (a,i,j))
	B_set (b,j,i);
  return b;
}

/* This code tries to find a few asymmetries in the bitmatrix, and to
   reflect and rotate them in a canonical way.  */

void B_init()
{
  int n;
  int a,b,c,d;

  /* Initialize array for B_choose_symmetry */
  for (n = 0; n < 16; n++) {
    a = (n&(1<<3)) ? 1 : 0;
    b = (n&(1<<2)) ? 1 : 0;
    c = (n&(1<<1)) ? 1 : 0;
    d = (n&(1<<0)) ? 1 : 0;
    symmetries_for_4[n] =
      (a+b > c+d ? DO_NOT_REFLECT_X_AXIS : 0) |
      (a+b < c+d ? DO_REFLECT_X_AXIS: 0) |
      (a+c > b+d ? DO_NOT_REFLECT_Y_AXIS : 0) |
      (a+c < b+d ? DO_REFLECT_Y_AXIS : 0) |
      ( a &&  b && !c && !d ? DO_NOT_TRANSPOSE : 0) |
      (!a && !b &&  c &&  d ? DO_NOT_TRANSPOSE : 0) |
      ( a && !b &&  c && !d ? DO_TRANSPOSE : 0) |
      (!a &&  b && !c &&  d ? DO_TRANSPOSE : 0);
  }
}


int B_choose_symmetry (B_type a, int symmetries_permitted) {
  int remaining = symmetries_permitted;
  int do_these = 0;
  int i,j;
  int x=B_x(a)-1, y=B_y(a)-1;
  int t1, t2, set, values;

  if (x > y) {
    remaining &= -1 ^ TRANSPOSE;
  } else if (x < y) {
    remaining &= -1 ^ TRANSPOSE;
    do_these |= TRANSPOSE;
  }
  for (i=0; i<(x+1)/2; i++) {
    for (j=0; j<(y+1)/2; j++) {
      t1 =(B_get(a,  i,  j) ? (1<<3) : 0)
	| (B_get(a,x-i,  j) ? (1<<2) : 0)
	| (B_get(a,  i,y-j) ? (1<<1) : 0)
	| (B_get(a,x-i,y-j) ? (1<<0) : 0);
      t2 = symmetries_for_4[t1];
      set = (t2 >> 3) & 7 & remaining;
      values = t2 & 7;
      remaining &= -1 ^ set;
      do_these |= (set & values);
      do_these &= -1 ^ (set & (-1 ^ values));
      if (remaining==0)
	return do_these;
    }
  }
  return do_these;
}

B_type B_symmetry (B_type a, int symmetries) {
  B_type b;
  int i,j,ii,jj;
  int x_start, y_start, x_adj, y_adj;
  int x=B_x(a), y=B_y(a);
  boolean transpose = symmetries & TRANSPOSE;
  
  if (symmetries & REFLECT_Y_AXIS) {
    x_adj = -1;
    x_start = x-1;
  } else {
    x_adj = 1;
    x_start = 0;
  }
  if (symmetries & REFLECT_X_AXIS) {
    y_adj = -1;
    y_start = y-1;
  } else {
    y_adj = 1;
    y_start = 0;
  }
  b = transpose ? B_make (y, x) : B_make (x, y);
  for (i=0,ii=x_start; i<x; i++, ii+=x_adj)
    for (j=0,jj=y_start; j<y; j++,jj+=y_adj)
      if (B_get (a,i,j))
	if (transpose) B_set (b,jj,ii);
	else B_set (b,ii,jj);
  return b;
}

/* Mike Earnsts' way of doing a similar thing. */

/* More orientation cleverness for efficient hashing:  If both of the
bottom corners are occupied, make sure that both of the top ones are too.
If only one of the top corners is occupied, make sure that the left-hand
one is.  */

/* I ignore edges because I started out doing so and it's simplest to
continue doing so.  I have no clear idea whether orienting according to
which edges are present would be better than this strategy, and no
particular reason to suspect it would be significantly so.  Also because
this is an operation on bitmatrices, not Konane representations. */

/* Returns an integer:  1s bit is set if it was flipped horizontally.
                        2s bit is set if it was flipped vertically. */
int B_flip (B_type a)
{ boolean temp;
  int i, j, k;
  boolean result = 0;
  
  /* perhaps consolidate these two flips */
  if (B_get(a,0,B_y(a)-1) & !B_get(a,0,0))
    /* flip horizontally */
    { for (i=0; i<B_x(a); i++)
	for (j=0,k=B_y(a)-1; j<k; j++,k--)
	  { temp = B_get(a,i,j);
	    if (B_get (a,i,k))
	      B_set (a,i,j);
	    else
	      B_clear (a,i,j);
	    if (temp)
	      B_set (a,i,k);
	    else
	      B_clear (a,i,k); }
      result = result+1; }
  if (B_get(a,B_x(a)-1,0) & B_get(a,B_x(a)-1,B_y(a)-1)
      & !(B_get(a,0,0) & B_get(a,0,B_y(a)-1)))
    /* flip vertically */
    { for (j=0; j<B_y(a); j++)
	for (i=0,k=B_x(a)-1; i<k; i++,k--)
	  { temp = B_get(a,i,j);
	    if (B_get (a,k,j))
	      B_set (a,i,j);
	    else
	      B_clear (a,i,j);
	    if (temp)
	      B_set (a,k,j);
	    else
	      B_clear (a,k,j); }
      result = result+2; }
  return result;
}


/* Returns number of first row with a stone, or -1 if no row has a stone. */
int B_min_occupied_x (B_type a)
{
  int i, j, xlimit, ylimit;

  xlimit = B_x(a);
  ylimit = B_y(a);

  for (i=0; i < xlimit; i++)
    for (j=0; j < ylimit; j++)
      if (B_get (a,i,j))
	return i;

  return -1;
}

/* Returns number of last row with a stone, or -1 if no row has a stone. */
int B_max_occupied_x (B_type a)
{
  int i, j, ylimit;

  ylimit = B_y(a);

  for (i=B_x(a)-1; i >= 0; i--)
    for (j=0; j < ylimit; j++)
      if (B_get (a,i,j))
	return i;

  return -1;
}


/* Returns number of first column with a stone, or -1 if no column has a stone. */
int B_min_occupied_y (B_type a)
{
  int i, j, xlimit, ylimit;

  xlimit = B_x(a);
  ylimit = B_y(a);


  for (j=0; j < ylimit; j++)
    for (i=0; i < xlimit; i++)
      if (B_get (a,i,j))
	return j;

  return -1;
}


/* Returns number of last column with a stone, or -1 if no column has a stone. */
int B_max_occupied_y (B_type a)
{
  int i, j, xlimit;

  xlimit = B_x(a);


  for (j=B_y(a)-1; j >= 0; j++)
    for (i=0; i < xlimit; i++)
      if (B_get (a,i,j))
	return j;

  return -1;
}

void B_stat ()
{
    printf ("B_matrices: made %d, freed %d, extras %d\n",
	    count_B_makes, count_B_frees, count_B_makes - count_B_frees);
}
