/* Copyright (C) David Wolfe, 1990.  All rights reserved. */

/* Maintains an 2 dimensional bit array */

/* Maximum dimension of 255 is determined by B_to_list().  This is
   used to save space in the Hash Table */

/* Note that many of the procedures are macros, which expand arguments
   more than once.  So you cannot, for instance, say B_set(a,i,j++) */

typedef struct B_struct {
	int x;
	int y;
	unsigned *contents;
} *B_type;

#define SpOt(a,i,j) ((i) * (a)->y + (j))
#define ByTe(a,i,j) ((a)->contents + (SpOt(a,i,j) / INT_SIZE))
#define BiT(a,i,j) (SpOt(a,i,j) % INT_SIZE)

#define B_x(a) ((a)->x)
#define B_y(a) ((a)->y)
#define B_get(a,i,j)   (((*ByTe(a,i,j) &   (1<<BiT(a,i,j))) != 0) && (i) < B_x(a) && (j) < B_y(a) && (i) >= 0 && (j) >= 0)
#define B_set(a,i,j)   ( *ByTe(a,i,j) |=  (1<<BiT(a,i,j)))
#define B_clear(a,i,j) ( *ByTe(a,i,j) &= ~(1<<BiT(a,i,j)))
#define B_free(a) (free ((a)->contents), free (a), count_B_frees++)

#define REFLECT_X_AXIS (1<<0)
#define REFLECT_Y_AXIS (1<<1)
#define TRANSPOSE      (1<<2)

extern void		B_init(void);
extern int		count_B_frees;
extern int		count_B_makes;
extern B_type		B_make (int, int);
extern list_type	B_to_list (B_type);
extern B_type		B_pack (B_type);
extern void		B_printf (B_type);
extern void		B_printf_with_symbols (B_type, char *, char *); /* You can specify
							    strings to print for 0 and 1 */
extern B_type		B_copy (B_type);
extern B_type		B_copy_into (B_type, B_type);
extern B_type		B_copy_offset (B_type, int, int, int, int);
extern B_type		B_expand (B_type, int, int);
extern int		B_choose_symmetry (B_type, int symmetries);
extern B_type		B_symmetry (B_type, int symmetries);
			/* symmetries are given by bitwise-OR of */
			/* REFLECT_X_AXIS, REFLECT_Y_AXIS and TRANSPOSE */
extern B_type		B_transpose (B_type);
extern int		B_flip (B_type);
extern int		B_min_occupied_x(B_type);
extern int		B_max_occupied_x(B_type);
extern int		B_min_occupied_y(B_type);
extern int		B_max_occupied_y(B_type);
extern void		B_stat();
