/* Copyright (C) David Wolfe, 1990.  All rights reserved. */


#include <stdlib.h>
#include <stdio.h>
#include "externs.h"
#include "list.h"
#include "hash.h"
#include "rational.h"
#include "gameops.h"
#include "output.h"
#include "bitmatrx.h"
#include "domino.h"

#define HORIZONTAL (11)
#define VERTICAL (22)

/*
NOT IN USE:
#define DOMINO_CLEAN 34
#define DOMINO_KEEP 34
*/

int pos_x[]={-1,0,1,0};		/* the 4 directions */
int pos_y[]={0,-1,0,1};
game_type place_domino (B_type pos, int x, int y, int length, boolean direction) ;
B_type copy_connected_component_at (B_type, int, int);
void clear_cycles (B_type, B_type, B_type, list_type, int, int, int);
game_type try_to_apply_articulation_theorem (B_type);
game_type domino_connected_position (B_type);
game_type domino_reduced_position (B_type);
#define NO_DOMINO (-1)
#define ROOT (5)
#define APPROXIMATE FALSE

#define ANY_LENGTH_LEGAL 0

/* begin procedures added by Tom VanDeGrift */
list_type find_bound(B_type);
B_type copy_split_domino(B_type, int, int, int, int);
game_type do_bounds_match(B_type);
int B_size_of(B_type);
boolean can_left_win_going_first(B_type);
boolean split_check(B_type, boolean *);
boolean alpha_beta(B_type);
game_type print_winner (B_type pos);
typedef int outcome_type;
#define OUTCOME_ZERO 0
#define OUTCOME_INCOMPARABLE 1
#define OUTCOME_LEFT_WINS 2
#define OUTCOME_RIGHT_WINS 3
outcome_type who_wins(B_type);

int domino_size_limit=0;
int domino_impartial=FALSE;
int domino_min_legal_Left_length=2;
int domino_max_legal_Left_length=2;
int domino_min_legal_Right_length=2;
int domino_max_legal_Right_length=2;
#define small_enough(pos) (B_size_of(pos) <= domino_size_limit)
  
/* end procedures added by Tom VanDeGrift */


game_type domino_position (B_type pos)
{
    game_type g=zero;
    int i, j;
    B_type New;		/* New is capitalized to avoid
			   the name conflict with C++ */
    for (j=0; j < B_y(pos); j++)
	for (i=0; i < B_x(pos); i++)
	    if (B_get (pos, i, j)) {
		New = copy_connected_component_at (pos, i, j);
		g = plus (g, domino_connected_position (New));
	    }
    B_free (pos);
    return g;
}

boolean legal_to_place (B_type pos, int x, int y, int length, boolean direction) 
{
    int i;
    if (direction == HORIZONTAL) {
	if (y < 0 || y >= B_y(pos)) return FALSE;
	for (i=0; i < length; i++,x++) {
	    if (x < 0 || x >= B_x(pos)) return FALSE;
	    if (!B_get(pos, x, y)) return FALSE;
	}
    } else if (direction == VERTICAL) {
	if (x < 0 || x >= B_x(pos)) return FALSE;
	for (i=0; i < length; i++,y++) {
	    if (y < 0 || y >= B_y(pos)) return FALSE;
	    if (!B_get(pos, x, y)) return FALSE;
	}
    } else myerror ("Bad direction");
    return TRUE;
}

game_type place_domino (B_type pos, int x, int y, int length, boolean direction) 
{
    int i;
    B_type New;

    New = B_copy (pos);
    for (i=0; i < length; i++, (direction==HORIZONTAL ? x++ : y++))
	B_clear(New, x, y);
    return domino_position (New);
}

game_type domino_connected_position (B_type pos)
{
    int symmetry;
    game_type g;
    B_type New;

    pos = B_pack(pos);
    symmetry = B_choose_symmetry (pos, REFLECT_X_AXIS | REFLECT_Y_AXIS | TRANSPOSE);
    New = B_symmetry (pos, symmetry);
    g = domino_reduced_position (New);
    B_free (pos);
    if (symmetry & TRANSPOSE)
	return negative (g);
    else
	return g;
}

game_type domino_reduced_position (B_type pos)
{
    int i, j, key, keep_key, length;
    list_type left, right, lhash;
    game_type g;
    int x=B_x(pos),y=B_y(pos);
/*  int count_squares=0; */
	
/*	if ((g = try_to_apply_articulation_theorem (pos)) != NO_GAME)
	return g;
*/
    lhash = B_to_list (pos);
    key = (list_nth(lhash, 1) << 8) | DOMINO_KEY;
    keep_key = (list_nth(lhash, 1) << 8) | DOMINO_KEEP_KEY;
    list_delete_nth (lhash, 1);
    if (hash_test (key,  lhash))
	return B_free (pos), list_free (lhash), hash_get_last();
/*  else if (hash_test (keep_key,  lhash))
	return B_free (pos), list_free (lhash), hash_get_last(); */
    left = list_make();
    right = list_make();
    for (i = 0; i < x; i++)
	for (j = 0; j < y; j++)
	    for (length = domino_min_legal_Left_length;
		 length <= domino_max_legal_Left_length
		     && legal_to_place (pos, i, j, length, VERTICAL);
		 length++)
		list_insert (left, place_domino (pos, i, j, length, VERTICAL));
    for (i = 0; i < x; i++)
	for (j = 0; j < y; j++)
	    for (length = domino_min_legal_Right_length;
		 length <= domino_max_legal_Right_length
		     && legal_to_place (pos, i, j, length, HORIZONTAL);
		 length++)
		list_insert (right, place_domino (pos, i, j, length, HORIZONTAL));
    if (domino_impartial) {
	g = make_nonpartisan (list_union (left, right));
	list_free (left);
	list_free (right);
    } else   
	g = make (left, right);
    hash_put (key, lhash, g);
/*  hash_put (((count_squares >= DOMINO_KEEP) ? keep_key : key),
	      lhash, g);
	      */
/*  if (count_squares >= DOMINO_CLEAN) {
	if (count_squares >= DOMINO_KEEP) {
	    B_printf (pos);
	    game_printf (g);
	    stat();
	}
	hash_clear(EXPENDABLE_KEYS,1024);
    }
*/
    B_free (pos);
    return g;
}

void copy_connected_component_at_ (B_type old, B_type New, int x, int y)
{
    int i,j,n;

    B_clear (old, x, y);
    B_set (New, x, y);
    for (n = 0; n <= 3; n++) {
	i = x + pos_x[n];
	j = y + pos_y[n];
	if (B_get (old,i,j))
	    copy_connected_component_at_ (old, New, i, j);
    }
}

B_type copy_connected_component_at (B_type old, int x, int y)
{
    B_type New;
    
    New = B_make (B_x (old), B_y (old));
    copy_connected_component_at_ (old, New, x, y);
    return New;
}

#undef max
#define max(a, b)	((a) > (b) ? (a) : (b))

B_type domino_input(void)
{
    B_type pos, New;
    char s[MAXBUFF];
    int i=0, j;
    
    pos = B_make(10,10);
    printf ("Enter domineering position, then hit <return> twice.\n");
    for (j=0; input_line(s) != NULL && *s != '\0';j++)
	for (i=0; *(s+i) != '\0'; i++) {
	    if (i >= B_x (pos) || j >= B_y (pos)) {
		New = B_make (10+max(i,B_x(pos)), 10+max(j, B_y(pos)));
		B_copy_into (New, pos);
		B_free (pos);
		pos = New;
	    }
	    if (*(s+i) != ' ') B_set (pos, i, j);
	}
    return B_pack(pos);
}

game_type domino ()
{
    B_type pos;
    game_type g;
	
    pos = domino_input();
    if((domino_size_limit < 1) || small_enough(pos)) {  
	g = domino_position(pos); 
	return g; 
    } 
    switch (who_wins(pos))
    {
    case OUTCOME_ZERO:
	printf("Game is 0\n");
	break;
    case OUTCOME_INCOMPARABLE:
	printf("Game is Incomparable to 0\n");
	break;
    case OUTCOME_LEFT_WINS:
	printf("Game is Positive: Left Wins\n");
	break;
    case OUTCOME_RIGHT_WINS:
	printf("Game is Negative: Right Wins\n");
	break;
    }
    B_free (pos);
    return NO_GAME;
}

/* The reminder of this files contains code to apply the Theorem on
   Winning Ways page 137.  Unfortunately, it does not seem to improve
   the efficiency of the domineering algorithm, and so the call to the
   procedure try_to_apply_articulation_theorem() in
   domino_connected_position() above has been commented out.  Hence,
   the following code is not used. */

game_type try_to_apply_articulation_theorem (B_type pos)
    /* Theorem on Winning Ways page 137 */
{
    int i=0, j=0, dir;
    list_type l;
    B_type arts, blank;
    B_type temp_pos, copy_pos;
    game_type g, h;
  
    i=0, j=0;
    while (!B_get (pos, i, j)) {	/* Find just one square */
	if (j < B_y (pos) - 1) j++;
	else if (i < B_x (pos) - 1) j=0, i++;
	else return B_free (pos), zero;
    }
    blank = B_make (B_x (pos), B_y (pos));
    arts = B_copy (pos);
    clear_cycles (pos, arts, blank, l = list_make(), i, j, ROOT);
    list_free (l);
    B_free (blank);
    for (i=0; i<B_x(pos); i++) {
	for (j=0; j<B_y(pos); j++) {
	    if (B_get (arts, i,j)) for (dir = 0; dir<4; dir++) {
		if (    B_get (pos, i+pos_x[dir], j+pos_y[dir])
			&&  B_get (pos, i-pos_x[dir], j-pos_y[dir])
			&& !B_get (pos, i+pos_y[dir], j+pos_x[dir])
			&& !B_get (pos, i-pos_y[dir], j-pos_x[dir])) {
		    copy_pos = B_copy (pos);
		    B_clear (copy_pos, i+pos_x[dir], j+pos_y[dir]);
		    temp_pos = copy_connected_component_at (copy_pos, i, j);
		    g = domino_connected_position (temp_pos);
	  
		    B_copy_into (copy_pos, pos);
		    B_clear (copy_pos, i, j);
		    temp_pos = copy_connected_component_at (copy_pos, i-pos_x[dir], j-pos_y[dir]);
		    h = domino_connected_position (temp_pos);
					
		    if (eq (g,h)) {
			B_copy_into (copy_pos, pos);
			B_clear (copy_pos, i-pos_x[dir], j-pos_y[dir]);
			temp_pos = copy_connected_component_at (copy_pos, i, j);
			h = domino_connected_position (temp_pos);
			B_free (copy_pos);
			B_free (arts);
			return plus (g,h);
		    }
		    B_free (copy_pos);
		}
	    }
	}
    }
    B_free (arts);
    return NO_GAME;
}

void clear_cycles (B_type pos, B_type arts, B_type explored,
		   list_type path, int x, int y, int parent_dir) {
    /* Assumes positions is connected.  Uses depth first search, */
    /* updating the path to root.  For each back edge in the dfs, */
    /* zero out all entries in arts along the induced cycle. */
    /* What's left should be the articulation points, i.e., those */
    /* boxes which disconnect the position when removed. */
    int i, j, n, ii, jj=0, dir;
    list_type l;
    boolean backedge;

    B_set (explored, x, y);
    list_prepend (path, y);	/* path denoted (x1, y1, x2, y2, x3, y3, ...) */
    list_prepend (path, x);
    for (n=0; n<(parent_dir == ROOT?4:3); n++) {
	dir = (parent_dir+3+n) % 4;
	i = x + pos_x[dir];
	j = y + pos_y[dir];
	if (B_get (pos, i, j)) {
	    if (B_get (explored, i, j)) {
		/* If it's a back edge, clear cycle */
		for (l = list_start (path), backedge = FALSE; l != NIL && !backedge;) {
		    ii = list_pick (l);
		    l = list_rest (l);
		    jj = list_pick (l);
		    l = list_rest (l);
		    if (ii == i && jj == j)
			backedge = TRUE;
		}
		if (backedge)
		    for (l = list_start (path), ii=i-1 ; i != ii || j != jj;) {
			ii = list_pick (l);
			l = list_rest (l);
			jj = list_pick (l);
			l = list_rest (l);
			B_clear (arts, ii, jj);
		    }
	    } else
		/* forward edge --- continue dfs */
		clear_cycles (pos, arts, explored, path, i, j, dir);
	}
    }
    list_delete_nth (path, 1);
    list_delete_nth (path, 1);
    return;
}

/* Procedures added by Tom VanDeGrift in May of 1997 */

/* determing the number of empty positions in the domino position */
int B_size_of(B_type pos) {
    int num = 0;
    int i, j;
  
    for(i = 0; i < B_x(pos); i++) 
	for(j = 0; j < B_y(pos); j++)
	    if (B_get(pos, i, j)) num++;
    return num;
}


list_type find_bound(B_type pos) {

    game_type g;
    list_type bounding_values;
    list_type bounds;
    list_type left, right, tmp;
    int x_dom, y_dom;
    B_type B_tmp;
    int size;

    /* if the number of squares in the domino position is <= some number
     * compute the value the starndard way else split the position
     * and compute the values of these and see if we get a tignt bound
     */

    bounding_values = list_make();
    x_dom = B_x(pos);
    y_dom = B_y(pos);

    if(small_enough(pos) || (x_dom <= 2) || (y_dom <= 2)) { 
	g = domino_position(B_copy(pos));
	list_insert(bounding_values, APPROXIMATE ? project(g) : g);
	return bounding_values;
    }

    for(size = (x_dom-1)/2; size <= x_dom/2 + 1 ;size++){
	left = find_bound (B_tmp = copy_split_domino(pos, 0, 0, size, y_dom));
	B_free (B_tmp);
	right = find_bound (B_tmp = copy_split_domino(pos, size, 0, x_dom, y_dom));
	B_free (B_tmp);
	remove_dominated_options(left, le);
	remove_dominated_options(right, le);
	bounds = list_mapcomb(plus, left, right);
	remove_dominated_options(bounds, le);
	bounding_values = list_union(tmp = bounding_values, bounds);
	list_free (tmp);
	list_free (bounds);
	list_free(left);
	list_free(right);
    }
    remove_dominated_options (bounding_values, le);
    return bounding_values;
}

B_type copy_split_domino (B_type old, int start_x, int start_y,
			  int end_x, int end_y)
{
    int i, j;
    B_type New = B_make (end_x - start_x, end_y - start_y);

    for (i=start_x; i<end_x; i++)
	for (j=start_y; j<end_y; j++)
	    if (B_get (old,i,j))
		B_set (New,i - start_x,j - start_y);
	    else
		B_clear (New,i - start_x,j - start_y);
    return B_pack(New);
}


outcome_type who_wins(B_type pos) {
  
    B_type New, right;
  
    boolean left_wins, right_wins;
  
    New = B_pack(B_copy (pos));
  
    right = B_transpose(New);
    left_wins = can_left_win_going_first(New);
    right_wins = can_left_win_going_first(right);
    B_free(right);
    B_free(New);

    if (left_wins && right_wins)
	return OUTCOME_INCOMPARABLE;
    else if (left_wins)
	return OUTCOME_LEFT_WINS;
    else if (right_wins)
	return OUTCOME_RIGHT_WINS;
    else
	return OUTCOME_ZERO;
}

boolean can_left_win_going_first (B_type pos)
{
    boolean result;

    if(small_enough(pos) || (B_x(pos) <= 2) || B_y(pos) <= 2)
	result = (!le (domino_position(B_copy(pos)),zero));
    else if (split_check (pos, &result))
	/* Do nothing */;
    else
	result = alpha_beta(pos);
    return result;
}

boolean can_right_win_going_first (B_type pos)
{
    B_type trans;
    boolean result;
    
    trans = B_transpose (pos);
    result = can_left_win_going_first (trans);
    B_free (trans);
    return result;
}

boolean split_check(B_type pos, boolean *result)
{
    list_type upper_bound, temp_bound, lower_bound, l;
    B_type B_tmp;

    upper_bound = find_bound(B_tmp = B_copy(pos)); 
    B_free (B_tmp);
    for (l = list_start (upper_bound); l != NIL; l = list_rest(l))
	if (APPROXIMATE ? Q_lt (left_stop (list_pick(l)), Q_0) : le(list_pick(l), zero)) {
	    *result = FALSE;
	    list_free (upper_bound);
	    return TRUE;
	}
    list_free (upper_bound);
    temp_bound = find_bound(B_tmp = B_transpose(pos));
    lower_bound = list_map(negative, temp_bound);  
    B_free (B_tmp);
    list_free (temp_bound);
    for (l = list_start (lower_bound); l != NIL; l = list_rest(l))
	if (APPROXIMATE ? Q_gt (left_stop (list_pick(l)), Q_0) : !le(list_pick(l), zero)) {
	    *result = TRUE;
	    list_free (lower_bound);
	    return TRUE;
	}
    list_free (lower_bound);
    return FALSE;
}

boolean alpha_beta(B_type pos) {
    B_type trans;
    boolean right_wins;
    int i, j;
  
    pos = B_pack(B_copy(pos));
    trans = B_transpose(pos);
    for (i = 0; i < B_x(trans) - 1; i++)
	for (j = 0; j < B_y(trans); j++)
	    if (B_get (trans, i, j) && B_get(trans, i + 1, j)) {
		B_clear (trans, i, j);
		B_clear (trans, i + 1, j);
		right_wins = can_left_win_going_first(trans);
		if(FALSE == right_wins) {
		    B_free(pos);
		    B_free(trans);
		    return (TRUE);
		}
		B_set (trans, i, j);
		B_set (trans, i + 1, j);
	    }
    B_free(pos);
    B_free(trans);
    return (FALSE);
}

/* End procedures added by Tom VanDeGrift in May of 1997 */

void domino_printf (B_type pos)
{
    B_printf_with_symbols (pos, " ", "x");
}

void domino_winning_left_move (B_type pos) 
{
    int i,j;
    
    for (i = 0; i < B_x(pos); i++)
	for (j = 0; j < B_y(pos)-1; j++)
	    if (B_get (pos, i, j) && B_get(pos, i, j+1)) {
		B_clear (pos, i, j);
		B_clear (pos, i, j+1);
		if (!can_right_win_going_first(pos)) {
		    return;
		}
		B_set (pos, i, j);
		B_set (pos, i, j+1);
	    }
    myerror ("Can't win from this domineering position");
}

boolean domino_desparation_left_move (B_type pos) 
{
    int i,j;
    
    for (i = 0; i < B_x(pos); i++)
	for (j = 0; j < B_y(pos)-1; j++)
	    if (B_get (pos, i, j) && B_get(pos, i, j+1)) {
		B_clear (pos, i, j);
		B_clear (pos, i, j+1);
		return TRUE; /* Legal move played */
	    }
    return FALSE; /* Lost position */
}
    
void domino_play () 
{
    B_type pos;
    boolean can_play = FALSE;

    pos = domino_input();
    switch (who_wins (pos)) {
    case OUTCOME_INCOMPARABLE:
    case OUTCOME_LEFT_WINS:
	can_play = TRUE;
	domino_winning_left_move (pos);
	break;
    case OUTCOME_RIGHT_WINS:
    case OUTCOME_ZERO:
	can_play = domino_desparation_left_move (pos);
	break;
    }
    if (can_play)
	domino_printf (pos);
    B_free (pos);
}
