/* Copyright (C) David Wolfe, 1990.  All rights reserved. */

#include <string.h>

#ifndef MAX_GAMES
#ifdef __linux__
#define MAX_GAMES (200000)
#define TABLE_SIZE (1000000)
#else
#define MAX_GAMES (50000)
#define TABLE_SIZE (200000)
#endif
#endif

#ifdef __NeXT__
#ifdef __cplusplus
#undef __cplusplus
#endif
#ifdef c_plusplus
#undef c_plusplus
#endif
#endif

#define INT_SIZE (32)

#define FALSE (0)
#define TRUE (1)

#define MAXBUFF (300)
#define NO_LIMIT (-1)
#define NO_PERIOD (-1)

/* Mostly print options */
#define INTEGERS_FLAG (1<<0)
#define NUMBERS_FLAG (1<<1)
#define STARS_FLAG (1<<2)
#define UPS_FLAG (1<<3)
#define TINYS_FLAG (1<<4)
#define ZEROS_FLAG (1<<5)
#define ONS_FLAG (1<<6)
/* Substitute variable names for games when a variable is available */
#define VARS_FLAG (1<<7)
/* However, if you're only printing the variable name, expand it */
#define EXPAND_FLAG (1<<8)
#define PROMPT_FLAG (1<<9)
/* check if g = h.1* */
#define MAP_FLAG (1<<10)
#define EVALUATE_FLAG (1<<12)
#define ECHO_INPUT_FLAG (1<<15)
/* reinitialize hash table between commands */
#define CLEAN_FLAG (1<<11)
/* output formulas in TeX */
#define TEX_FLAG (1<<13)
/* trace evaluation */
#define TRACE_FLAG (1<<10)
/* evaluate Konane positions using modern rules */
#define MODERN_KONANE_FLAG (1<<16)
#define DEFAULT_FLAG (INTEGERS_FLAG|NUMBERS_FLAG|STARS_FLAG|UPS_FLAG|TINYS_FLAG|\
                  VARS_FLAG|EXPAND_FLAG|PROMPT_FLAG|ZEROS_FLAG)
#define ALL_FLAG     (INTEGERS_FLAG|NUMBERS_FLAG|STARS_FLAG|UPS_FLAG|TINYS_FLAG|\
                  VARS_FLAG|EXPAND_FLAG|PROMPT_FLAG|ZEROS_FLAG|ONS_FLAG)

/* hash table entries */
/* Note that domino.c has been optimized and depends upon
   the first 16 bits of these KEYS being unused */
#define UNEXPENDABLE_KEYS	(1 << 5)
#define EXPENDABLE_KEYS		(1 << 6)
#define OTHER_KEYS		(1 << 7)
#define KEY_PORTION		(31)
#define SIMPLIFY_KEY		( 0 | UNEXPENDABLE_KEYS)
#define PLUS_KEY		( 1 | EXPENDABLE_KEYS)
#define NEGATIVE_KEY		( 2 | EXPENDABLE_KEYS)
#define GE_KEY			( 3 | EXPENDABLE_KEYS)
#define NUM_UP_STAR_KEY		( 4 | EXPENDABLE_KEYS)
#define COOL_KEY		( 5 | EXPENDABLE_KEYS)
#define TEMPERATURE_KEY		( 6 | EXPENDABLE_KEYS)
#define FREEZE_KEY		( 7 | EXPENDABLE_KEYS)
#define HEAT_KEY		( 8 | EXPENDABLE_KEYS)
#define MULT_KEY		( 9 | EXPENDABLE_KEYS)
#define AW_KEY			(10 | EXPENDABLE_KEYS)
#define VAR_KEY			(11 | UNEXPENDABLE_KEYS)
#define NUM_EPSILON_KEY		(12 | UNEXPENDABLE_KEYS)
#define DOMINO_KEY		(13 | EXPENDABLE_KEYS)
#define DOMINO_KEEP_KEY		(14 | UNEXPENDABLE_KEYS)
/* Hashtable entry encodes ancient vs. modern, so just one key. */
#define KONANE_KEY		(15 | EXPENDABLE_KEYS)
#define TOADS_KEY		(16 | EXPENDABLE_KEYS)
#define SUBTRACTION_KEY		(17 | EXPENDABLE_KEYS)
#define INFINITY		((~0) & (~(1<<31)))
/* These indicate what to do when hash table needs space.
   OTHER_KEYS are sometimes expendable, depending on context */

/* ARGUMENT_SYMBOL is the internal string representation of the argument to 
   cool_function and warm_function. */
#define ARGUMENT_SYMBOL "\1"

typedef int boolean;
typedef unsigned game_type;
typedef game_type (* function_type)();
typedef boolean (* test_type) (game_type, game_type);

extern void		help();
extern void		stat();
extern void		control_C (int);	/* Trap control C's */
extern void		myerror (char *);
extern int		game_parse(char *);
extern char **		envp;
extern unsigned long	flags;

extern char help_file[];
/* extern void _exit(int), perror(const char *), execle(); */
/* extern int fork(), wait(), yyparse(), fflush() printf(const char *, ...) */
/* extern char *gets(); */
#ifdef __TURBOC__
extern void exit(int), free(void *);
extern long coreleft(void);
#endif

extern int tlevel;	/* Trace level */
extern boolean control_C_was_hit;
extern boolean delay_control_C;
#define disable_interrupts (delay_control_C=TRUE)
/* The ,0 is so both arms of the conditional return the same type. */
#define enable_interrupts (delay_control_C=FALSE, control_C_was_hit ? (control_C(SIGINT), 0) : 0)
extern char *input_line(char *);
