/* Copyright (C) David Wolfe, 1990.  All rights reserved. */

#include <sys/wait.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include "externs.h"
#include "rational.h"
#include "list.h"
#include "hash.h"
#include "bitmatrx.h"
#include "gameops.h"
#include "output.h"
#include <setjmp.h>
#include <signal.h>
#include <time.h>

/* Declarations for fork() and execle */
#ifdef __GNUC__
#ifndef __NeXT__
#include <unistd.h>
#endif
#endif

/* Prevent warning if -Wstrict-prototypes.  */
#ifdef __GNUC__
int yyparse (void);
#endif

char			help_file[MAXBUFF];
unsigned long		flags = DEFAULT_FLAG;
char 			**envp;
jmp_buf			env;
int tlevel = 0;
time_t starting_time;

void myerror(char *s) {
	if (strlen (s) >= 5)
		printf ("%s\n", s);
	else
		perror (s);
}

void help() {
#ifdef LIB_DIR
	char help_filename[MAXBUFF];
	FILE *infile;

	strcpy (help_filename, LIB_DIR);
	if (strlen (help_file) > 0) {
		strcat (help_filename, "H_");
		strcat (help_filename, help_file);
	}
	else
		strcat (help_filename, "HELP");
	if ((infile = fopen (help_filename, "r")) == NULL)
		if (strlen (help_file) > 1)
			printf ("No help available on that topic.\n");
		else
			printf ("Help seems to be broken on this system.  Sorry.\n");
	else {
#ifdef PRINT_COMMAND		
		fclose (infile);
		switch (fork ()) {
		case -1:
			myerror ("Sorry, couldn't fork");
			break;
		case 0:
			execle (PRINT_COMMAND, PRINT_COMMAND, help_filename, 0, envp);
			myerror ("Sorry, couldn't exec");
			_exit (-1);
		default:
			if (wait(0) < 0)
				myerror ("wait in help");
		}
#else
		{
			char line[MAXBUFF];
			while (fgets (line, MAXBUFF, infile) != '\0')
				printf ("%s", line);
			fclose (infile);
		}
#endif
	}
#else
	printf ("Help seems to be broken on this system.  Sorry\n");
#endif
}
	


/* NOT NEEDED IF USING GNU */
#if 0
extern char *yysptr;
extern char yysbuf[];
#endif
int game_parse (char *s) {
	parser_input = s;
	parser_output = NO_GAME;
	if (flags & CLEAN_FLAG) clean();
#if 0
	yysptr = yysbuf;	/* Initialize lex's unput buffer to empty in case 
				 * last call to yyparse() bombed. */
#endif
	if (yyparse()) return NO_GAME;
	else return parser_output;
}

extern unsigned long game_stat(void);
extern unsigned long hash_stat(void);
extern unsigned long list_stat(void);
boolean control_C_was_hit=FALSE;
boolean delay_control_C=FALSE;

void stat (void) {
	printf ("Elapsed time (in seconds): %lu\n", time(0) - starting_time);
#ifdef __TURBOC__
	printf ("total lists: %lu; coreleft %lu\n",
		hash_stat() + game_stat() + list_stat(), coreleft());
#else
	B_stat();
	printf ("total list cells found: %lu\n",
		hash_stat() + game_stat() + list_stat());
#endif
}

void control_C (int sig) {
	char s[MAXBUFF];

	signal(SIGINT, SIG_IGN);
	if (delay_control_C) {
		control_C_was_hit = TRUE;
		return;
	}
	control_C_was_hit = FALSE;
	while (1) {
		printf ("\n");
		printf ("q\tQuit\n");
		printf ("a\tAbort current computation only\n");
		printf ("c\tContinue execution\n");
		printf ("s\tStatus --- debugging information\n");
		while (1) {
		  printf ("Choose from the above: ");
		  switch (*gets(s)) {
		  case 'q':
		    exit(0);
		  case 'a':
		    signal(SIGINT, control_C);
		    longjmp (env, 1);
		  case 'c':
		    signal(SIGINT, control_C);
		    return;
		  case 's':
		    stat();
		    break;
		  default:
		    continue;
		  }
		  break;
		}
	}
}

char *input_line (char *input) {
	char *s;
	boolean test;

	if (fgets (input, MAXBUFF, stdin) == NULL) exit(0);
	for (s=input; *s!='\n' && *s != '\0';s++);
	if (*s=='\n')
		*s = '\0';
	else {
		/* Input is too long.  Eat it up and print an error message. */
		for (test=TRUE;test;) {
			if (fgets (input, MAXBUFF, stdin) == NULL) exit(0);
			for (s=input; *s!='\n' && *s != '\0';s++);
			if (*s == '\n')
				test=FALSE;
		}
		*input = '\0';
		myerror ("Input line to long");
	}
	return input;
}

void games (void) {
	char input[MAXBUFF];
	game_type g;

	starting_time = time (0);
	printf ("Type 'help' and 'help help'\n");
	if (setjmp (env) == 0); /* ignore return code for now */
	signal(SIGINT, control_C);
	for (;;) {
		if (flags & PROMPT_FLAG) printf ("> ");
		if (input_line (input) == NULL || !strncmp (input,"quit",4) || !strncmp (input,"exit",4)) {
			printf ("\n");
			exit(0);	/* eof */
		}
		if (flags & ECHO_INPUT_FLAG) printf ("%60s => ", input);
		fflush (stdout);
		strcat (input, "\n");
		g = game_parse (input);
		if (g) set_var (dummy_var = make_var ("$"), g);
		if (g) game_printf (g);
		fflush (stdout);
	}
}

