/* Copyright (C) David Wolfe, 1990.  All rights reserved. */
/* Includes mostly code written by David Moews */

%{
#include "externs.h"
#include "rational.h"
#include "list.h"
#include "gameops.h"
#include "output.h"
#include "bitmatrx.h"
#include "domino.h"
#include "konane.h"
#include "toads.h"
#include "subtraction.h"

#define vsl(a,b) make(list_copy_1((a)), (b))
#define vsr(a,b) make((a), list_copy_1((b)))

boolean non_syn_error = FALSE;
void yyerror ();
%}

%start line

%union {
	int ival;
	Q_type qval;
	list_type lval;
	game_type gval;
        var_type vval;
}

%type <vval> vs
%type <ival> mask invmask
%type <lval> l nl
%type <gval> num atom atoms numatoms term
%type <gval> g gt gts line gop gterm
%type <gval> gp		/* plain g (w/o bars in outer level) */
%type <gval> gb gb1 gb2 gb3 gb4

%token <ival> NUM ZEROS BARZEROS ONNUM
%token <vval> S
%token ENDOFSTRING
%token FREEZE CHILL MEAN TEMPERATURE AW ON OFF PROJECT COOL_BY_STAR
%token HELP RESET STAT
%token INTEGERS NUMBERS STARS UPS TINYS ONS VARS EXPAND ALL LIMIT IMPARTIAL EVALUATE PROMPT ECHO_ CLEAN TEX ZEROS ANCIENT MODERN MAP PLAY
%token BAR BARR BARRR BARRRR BARRRRR NO
%token MINUS_K MINUS_k
%token TINY MINY
%token UNKNOWN
%token DOMINO LENGTH ALL KONANE AKONANE MKONANE TOADS SUBTRACTION
%token SALTUS PERIOD TRACE

%%
line	: ENDOFSTRING /*empty*/	{ }
	| error ENDOFSTRING	{if (!non_syn_error)
    				    myerror("Syntax error.");
                                 non_syn_error = FALSE;
				 yyerrok;
   				}
	| g ENDOFSTRING		{parser_output = $1; YYACCEPT;}
	| g '?' g ENDOFSTRING	{(void)printf("%s\n", compare($1, $3));
					parser_output = NO_GAME;}
	| TRACE NUM ENDOFSTRING	{flags |= TRACE_FLAG; tlevel = $2;}
	| mask ENDOFSTRING	{flags |= $1;}
	| NO mask ENDOFSTRING 	{flags &= ~$2;}
	| invmask ENDOFSTRING 	{flags &= ~$1;}
	| HELP ENDOFSTRING	{(void) help();}
	| STAT ENDOFSTRING	{stat();}
	| RESET ENDOFSTRING	{reinit();}
	| DOMINO LENGTH ALL ENDOFSTRING {
	    domino_min_legal_Left_length = domino_min_legal_Right_length = 1;
	    domino_max_legal_Left_length = domino_max_legal_Right_length = INFINITY;
	    clean();
	}
	| DOMINO LENGTH NUM ENDOFSTRING {
	    domino_min_legal_Left_length = domino_min_legal_Right_length = $3;
	    domino_max_legal_Left_length = domino_max_legal_Right_length = $3;
	    clean();
	}
        | DOMINO ENDOFSTRING	{parser_output = domino ();}
	| PLAY DOMINO ENDOFSTRING {domino_play();}
	| DOMINO LIMIT NUM ENDOFSTRING {domino_size_limit = $3;}
	| DOMINO LIMIT ENDOFSTRING {domino_size_limit = 0;}
	| DOMINO IMPARTIAL ENDOFSTRING	{domino_impartial = TRUE;
				 clean();}
	| KONANE nl ENDOFSTRING	{parser_output = konane ($2, flags & MODERN_KONANE_FLAG);}
	| AKONANE nl ENDOFSTRING	{parser_output = konane ($2, FALSE);}
	| MKONANE nl ENDOFSTRING	{parser_output = konane ($2, TRUE);}
        | TOADS ENDOFSTRING	{parser_output = toads();}
	| SUBTRACTION nl ENDOFSTRING
				{parser_output = subtraction_table ($2,list_copy($2),NO_LIMIT);}
	| SUBTRACTION nl '(' NUM ')' ENDOFSTRING
				{parser_output = subtraction_table ($2,list_copy($2),$4);}
	| SUBTRACTION nl BAR nl ENDOFSTRING
				{parser_output = subtraction_table ($2,$4,NO_LIMIT);}
	| SUBTRACTION nl BAR nl '(' NUM ')' ENDOFSTRING
				{parser_output = subtraction_table ($2,$4,$6);}
	| SALTUS ENDOFSTRING	{parser_output = saltus; YYACCEPT;}
	| SALTUS g ENDOFSTRING	{saltus = $2;}
	| PERIOD ENDOFSTRING	{parser_output = g_int (period); YYACCEPT;}
	| PERIOD NUM ENDOFSTRING{period = $2;}
	;	

mask	: INTEGERS		{$$ = INTEGERS_FLAG;}
	| NUMBERS		{$$ = NUMBERS_FLAG;}
	| STARS			{$$ = STARS_FLAG;}
	| UPS			{$$ = UPS_FLAG;}
	| TINYS			{$$ = TINYS_FLAG;}
	| ZEROS			{$$ = ZEROS_FLAG;}
	| ONS			{$$ = ONS_FLAG;}
	| EXPAND		{$$ = EXPAND_FLAG;}
	| VARS			{$$ = VARS_FLAG;}
	| ALL			{$$ = ALL_FLAG;}
	| PROMPT		{$$ = PROMPT_FLAG;}
	| EVALUATE		{$$ = EVALUATE_FLAG;}
	| ECHO_			{$$ = ECHO_INPUT_FLAG;}
	| CLEAN			{$$ = CLEAN_FLAG;}
	| TEX			{$$ = TEX_FLAG;}
	| TRACE			{$$ = TRACE_FLAG;}
	| MODERN		{$$ = MODERN_KONANE_FLAG;}
	;

invmask : ANCIENT		{$$ = MODERN_KONANE_FLAG;}
  	;

nl	: 			{$$ = list_make();}
	| NUM nl		{list_prepend ($2, $1), $$ = $2;}
	 

l 	: gt					{$$ = list_copy_1($1);}
	| l ',' gt				{list_insert ($$ = $1, $3);}
	;

gb1	: l BAR l				{$$ = make($1, $3);}
	| l BAR gb1				{$$ = vsr($1, $3);}
	| ZEROS BAR gt				{$$ = zeros_vs ($3, $1);}
	| gt BARZEROS				{$$ = vs_zeros ($1, $2);}
	;

gb2	: gb1					{$$ = $1;}
	| l BARR l				{$$ = make($1, $3);}
	| l BARR gb2				{$$ = vsr($1, $3);}
	| gb1 BARR l				{$$ = vsl($1, $3);}
	| gb1 BARR gb2				{$$ = vs ($1, $3);}
	;

gb3	: gb2					{$$ = $1;}
	| l BARRR l				{$$ = make ($1, $3);}
	| l BARRR gb3				{$$ = vsr($1, $3);}
	| gb2 BARRR l				{$$ = vsl($1, $3);}
	| gb2 BARRR gb3				{$$ = vs ($1, $3);}
	;

gb4	: gb3					{$$ = $1;}
	| l BARRRR l				{$$ = make ($1, $3);}
	| l BARRRR gb4				{$$ = vsr($1, $3);}
	| gb3 BARRRR l				{$$ = vsl($1, $3);}
	| gb3 BARRRR gb4			{$$ = vs ($1, $3);}
	;

gb 	: gb4					{$$ = $1;}
	| l BARRRRR l				{$$ = make ($1, $3);}
	| l BARRRRR gb				{$$ = vsr($1, $3);}
	| gb4 BARRRRR l				{$$ = vsl($1, $3);}
	| gb4 BARRRRR gb			{$$ = vs ($1, $3);}
	;

g	: vs '=' g				{set_var($1, $$ = $3);}
	| gts
	;

gts	: gterm					{$$ = $1;}
	| gts '+' gterm				{$$ = plus($1, $3);}
	| gts '-' gterm				{$$ = minus($1, $3);}
	;

gterm	: gt					{$$ = $1;}
	| gb					{$$ = $1;}
	;

gt	: gop					{$$ = $1;}
	| gp					{$$ = $1;}
	| gp '[' g ']'				{$$ = cool ($1, mean ($3));}
	| gp PROJECT				{$$ = project ($1);}
	| gp COOL_BY_STAR				{$$ = cool_by_star ($1);}
	| gp '<' g '>'				{$$ = gexp ($1, Q_p (mean ($3)));}
	| gp '.' gp				{$$ = mult($1, $3);}
	| gp ONNUM				{$$ = on ($1, $2);}
	| gt gop				{$$ = plus ($1,$2);}
	;

gop	: FREEZE gp				{$$ = freeze ($2);}
	| '%' gp				{$$ = overheat ($2, NO_GAME, NO_GAME);}
	| '%' '<' g '>' gp			{$$ = heat ($5, $3);}
	| '%' '[' g ']' '<' g '>' gp		{$$ = overheat ($8, $3, $6);}
	| ON '[' g ']' '<' g '>'		{$$ = on ($3, Q_p (mean ($6)));}
	| OFF '[' g ']' '<' g '>'		{$$ = off ($3, Q_p (mean ($6)));}
	| MEAN gp				{$$ = num (mean ($2));}
	| TEMPERATURE gp			{$$ = num (temperature ($2));}
	| AW gp					{$$ = atomic_weight ($2);}
	;

num	: NUM					{$$ = num(Q_int($1));}
	| NUM '/' NUM				{$$ = num(Q($1, $3));}
	| '-' NUM '/' NUM			{$$ = num(Q(-$2, $4));}
	| NUM NUM '/' NUM			{$$ = num(Q_plus(Q_int ($1), Q($2, $4)));}
	| '-' NUM NUM '/' NUM			{$$ = num(Q_plus(Q_int (-$2), Q(-$3, $5)));}
	| '-' vs				{$$ = negative (get_var ($2));}
	| '-' NUM				{$$ = num(Q_int(-$2));}
	;

atom	: '*'					{$$ = star(1);}
	| '*' NUM				{$$ = star($2);}
	| '^'					{$$ = up(1);}
	| '^' NUM				{$$ = up($2);}
	| 'v'					{$$ = down(1);}
	| 'v' NUM				{$$ = down($2);}
	| TINY '[' g ']'			{$$ = tiny ($3);}
	| MINY '[' g ']'			{$$ = miny ($3);}
	| '{' g '}'				{$$ = $2;}
	| '(' g ')'				{$$ = $2;}
	;

atoms	: atom					{$$ = $1;}
	| atoms atom				{$$ = plus ($1, $2);}
	;

numatoms: atoms					{$$ = $1;}
	| num					{$$ = $1;}
	| num atoms				{$$ = plus($1, $2);}
	;

term	: numatoms				{$$ = $1;}
/*	| numatoms '%' numatoms			{$$ = plus ($1, mult ($3, one_star));}*/
	;

gp	: term					{$$ = $1;}
	| vs					{$$ = get_var($1);
						 list_free ($1);
						 if ($$ == NO_GAME) {
                                                   non_syn_error = TRUE;
                                                   YYERROR;
                                                 }
						}
	| '{' l BAR '}'				{$$ = make($2, list_make());}
	| '{' BAR l '}'				{$$ = make(list_make(), $3);}
	| '{' BAR '}'				{$$ = zero;}
	| '{' l BARR '}'			{$$ = make($2, list_make());}
	| '{' gb1 BARR '}'			{$$ = vsl($2, list_make());}
	| '{' BARR l '}'			{$$ = make(list_make(), $3);}
	| '{' BARR gb1 '}'			{$$ = vsr(list_make(), $3);}
	| '{' BARR '}'				{$$ = zero;}
	| '{' l BARRR '}'			{$$ = make($2, list_make());}
	| '{' gb2 BARRR '}'			{$$ = vsl($2, list_make());}
	| '{' BARRR l '}'			{$$ = make(list_make(), $3);}
	| '{' BARRR gb2 '}'			{$$ = vsr(list_make(), $3);}
	| '{' BARRR '}'				{$$ = zero;}
	| '{' l BARRRR '}'			{$$ = make($2, list_make());}
	| '{' gb3 BARRRR '}'			{$$ = vsl($2, list_make());}
	| '{' BARRRR l '}'			{$$ = make(list_make(), $3);}
	| '{' BARRRR gb3 '}'			{$$ = vsr(list_make(), $3);}
	| '{' BARRRR '}'			{$$ = zero;}
	| '{' l BARRRRR '}'			{$$ = make($2, list_make());}
	| '{' gb4 BARRRRR '}'			{$$ = vsl($2, list_make());}
	| '{' BARRRRR l '}'			{$$ = make(list_make(), $3);}
	| '{' BARRRRR gb4 '}'			{$$ = vsr(list_make(), $3);}
	| '{' BARRRRR '}'			{$$ = zero;}
	;

vs	: S					{$$ = $1;}
	;
%%
char *parser_input;
game_type parser_output;

void yyerror (s)
	char *s;
{
}

/* IF NOT USING GNU'S BISON, INCLUDE THE FOLLOWING
#include "parse.c"
*/
