/* Copyright (C) David Wolfe, 1990.  All rights reserved. */

#include <stdlib.h>
#include <signal.h>
#include <stdio.h>
#include "externs.h"
#include "list.h"
#include "hash.h"

#define HNIL ((hash_chain_type) 0)
#define Chain hash_last_chain

long unsigned		hash_count=0;
long unsigned		hash_cleanings=0;
hash_chain_type		hash_last_chain;
static hash_chain_type	hash_free_list = HNIL;
hash_chain_type		hash_table[TABLE_SIZE];

static hash_chain_type hash_get_cell (void) {
	hash_chain_type cell, p;
	int i;

	disable_interrupts;
	if (hash_free_list == HNIL) {
		if (!(hash_free_list = (hash_chain_type)
		      malloc(HASH_CELL_BLOCK_SIZE * sizeof (struct hash_chain_struct))))
			myerror("hash_get_cell: malloc couldn't free block");
		for (i = 1, p=hash_free_list; i < HASH_CELL_BLOCK_SIZE;i++, p++)
			p->next = p + 1;
		p->next = HNIL;
	}
	cell = hash_free_list;
	hash_free_list = hash_free_list -> next;
	hash_count++;
	enable_interrupts;
	return cell;
}

static void hash_free_cell (hash_chain_type list) {
	disable_interrupts;
	list -> next = hash_free_list;
	hash_free_list = list;
	hash_count--;
	enable_interrupts;
	return;
}

void hash_init (void) {
	int i;

	disable_interrupts;
	for (i=0;i<TABLE_SIZE;i++) hash_table[i] = HNIL;
	enable_interrupts;
	return;
}

void hash_clear(int keys, int rate) {
	int i;
	static int random=0;
	hash_chain_type chain, next;
	
	disable_interrupts;
	for (i = 0; i < TABLE_SIZE; i++) {
		while ((chain = hash_table[i]) != HNIL
		       && (chain->id & keys)
		       && (random = (random+373) & 1023) < rate) {
			hash_table[i] = chain->next;
			list_free (chain->key);
			hash_free_cell (chain);
		}
		if (chain != HNIL) {
			while ((next = chain->next) != HNIL)
				if (   (next->id & keys)
				    && (random = (random+373) & 1023) < rate) {
					chain->next = next->next;
					list_free (next->key);
					hash_free_cell (next);
				}
				else chain = next;
		}
	}
	enable_interrupts;
	return;
}

int hash_function (hash_id_type id, hash_key_type key) {
	unsigned long val;

	val = (A * id + B);
	for (key = list_start (key); key != NIL; key = list_rest (key))
		val = (A * (val + list_pick(key) + 1) + B);
	return val % TABLE_SIZE;
}

/*
int hash_function (hash_id_type id, hash_key_type key) {
	unsigned long val;

	val = (A * (id%M) + B) % M;
	for (key = list_start (key); key != NIL; key = list_rest (key))
		val = (A * (val + list_pick(key) + 1) + B) % M;
	return val % TABLE_SIZE;
}
*/
boolean hash_test (hash_id_type id, hash_key_type key) {
	for (Chain = hash_table[hash_function(id,key)]; Chain != HNIL; Chain = Chain->next)
		if ((Chain->id == id) && list_equal(Chain->key, key))
			return TRUE;
	return FALSE;
}
		
#define current_space() ((long) (hash_count * sizeof (struct hash_chain_struct) + list_count * sizeof (struct list_struct)))
void hash_put (hash_id_type id, hash_key_type key, hash_value_type value) {
	hash_chain_type New;
	int i;
	static long last_cleaned=0, last_warned=0;
		
	disable_interrupts;
	if (current_space() > SPACE_CLEAN)
		if (current_space() - last_cleaned > SPACE_INTERVAL) {
			hash_clear (EXPENDABLE_KEYS, 500);
			last_cleaned = current_space();
			hash_cleanings++;
		} else if (current_space() > SPACE_WARNING
			   && (current_space() - last_warned > SPACE_INTERVAL)) {
			myerror ("WARNING -- running out of space.");
			last_warned = current_space();
		}
	New = hash_get_cell();
	New -> value = value;
	New -> id = id;
	New -> key = key;

	i = hash_function(id, key);
	New -> next = hash_table[i];
	hash_table[i] = New;
	enable_interrupts;
	return;
}

hash_value_type hash_get (hash_id_type id, hash_key_type key) {
	for (Chain = hash_table[hash_function(id,key)]; Chain != HNIL; Chain = Chain->next)
		if ((Chain->id == id) && list_equal(Chain->key, key))
			return Chain->value;
	myerror ("hash_get: not in table");
	return 0;
}

void hash_delete (hash_id_type id, hash_key_type key) {
	hash_chain_type chain, next;

	disable_interrupts;
	chain = hash_table[hash_function(id,key)];
	if (chain == HNIL) return;
	else if ((chain->id == id) && list_equal(chain->key, key)) {
		hash_table[hash_function(id,key)] = chain -> next;
		list_free (chain->key);
		hash_free_cell (chain);
	} else while ((next = (chain->next)) != HNIL)
		if ((next->id == id) && list_equal(next->key, key)) {
			chain->next = next->next;
			list_free (next->key);
			hash_free_cell (next);
			return;
		} else chain = next;
	enable_interrupts;
	return;
}

unsigned long hash_stat (void) {
	unsigned long int n=0, c[30], j, k[30], i, l[30], x;
	hash_chain_type chain;

	for (i = 0; i < 30; i++) c[i]=k[i]=l[i]=0;
	for (i = 0; i < TABLE_SIZE; i++) {
		for (j=0, chain = hash_table[i]; chain != HNIL;j++, chain = chain->next) {
			x = list_length (chain->key) + 1;
			n = n + x;
			if (x>29) l[30]++;
			else l[x]++;
			(k[chain->id & 31])++;
		}
		if (j < 30)
			c[j]++;
		else
			printf ("Table entry %lu has %lu elements\n", i, j);
	}
	printf ("Distribution of hash table chains by length:\n");
	for (i = 0; i < 30; i++) printf ("%lu ", c[i]);
	printf ("\nDistribution of hash table usage by id:\n");
	for (i = 0; i < 30; i++) printf ("%lu ", k[i]);
	printf ("\nDistribution of hash table lists by length:\n");
	for (i = 0; i < 30; i++) printf ("%lu ", l[i]);
	printf ("\n");
	printf ("hash cleanings performed: %lu; hash table lists: %lu; ", hash_cleanings, n);
	return n;
}
