/* Copyright (C) David Wolfe, 1990.  All rights reserved. */

/* Hash table implemented using chaining.
 * NOTE:  There is only one hash table!  The first argument is an integer,
 * which will generally be a unique identifier to which hash you want,
 * simulating multiple hash tables (ie. the real key is constructed from
 * the pair    (id, key).  The intended purpose is for memoizing
 * functions, in which case the first argument should uniquely identify the
 * function.  All functions are written in a type indepenedent manner except
 * hash_fuction().
 */

#define HASH_CELL_BLOCK_SIZE (1000)	/* size of mallocs, when needed */
#ifndef TABLE_SIZE			/* in games program, defined in exters.h */
#define TABLE_SIZE (100000)
#endif
#define A ((unsigned long) 67343)
#define B ((unsigned long) 72763)

#ifdef IBMPC
#define SPACE_CLEAN	((long) 200000)
#define SPACE_INTERVAL	((long) 10000)
#define SPACE_WARNING	((long) 500000)
#else
/*
#define SPACE_CLEAN	((long) 20000000)
#define SPACE_INTERVAL	((long) 20000000)
#define SPACE_WARNING	((long) 40000000)
*/
#define SPACE_CLEAN	((long) 200000000)
#define SPACE_INTERVAL	((long) 200000000)
#define SPACE_WARNING	((long) 200000000)
#endif

typedef game_type hash_value_type;
typedef int hash_id_type;
typedef list_type hash_key_type;

typedef struct hash_chain_struct {
	hash_value_type value;
	hash_id_type id;
	hash_key_type key;
	struct hash_chain_struct *next;
} *hash_chain_type;

extern hash_chain_type	hash_table[TABLE_SIZE];
extern hash_chain_type	hash_last_chain;

extern void		hash_init (void);
extern void		hash_clear (int, int);
extern boolean		hash_test (hash_id_type, hash_key_type);
extern void		hash_put (hash_id_type, hash_key_type, hash_value_type);
extern hash_value_type	hash_get (hash_id_type, hash_key_type);
extern void		hash_delete (hash_id_type, hash_key_type);

#define hash_get_last() (hash_last_chain->value)
