/* konane.c */
/* by Michael Ernst, December 1991.  Last modified 1/20/92. */
/* Game-theoretically evaluate Konane positions.
   For use with David Wolfe's "games" program. */

/* Set to 0 (FALSE) to disable the efficiency hack for 1d positions. */
#define K1D_OPTIMIZE 1

#define KDEBUG 0
#define EXPAND_DEBUG 0
#define K1D_DEBUG 0
#define KDEBUG_DEPTH_LIMIT -1		/* -1 means no limit */

void kshowdepth(void);


/* Should I surround this with the __STDC__ test? */
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include "externs.h"
#include "list.h"
#include "hash.h"
#include "rational.h"
#include "gameops.h"
#include "output.h"
#include "bitmatrx.h"
/* #include "bitmatrix-mde.h" */
#include "konane.h"

/* Don't forget to free up space */

/* ***************************************************************************
   Debugging
   ********* */

int kdepth;

void kshowdepth (void)
{
  int i;

  for (i = 0; i < kdepth; i++)
    printf(">");
}


/* ***************************************************************************
   Abstraction
   *********** */

K_type K_make (int x, int y, int modern)
{
	K_type a;

	disable_interrupts;
	/* Was: a = K_make_empty(); */
	if (!(a = (K_type) malloc (sizeof (struct K_struct))))
		myerror ("K_make: malloc couldn't free block");
	a->polarity = 0;
	a->modern = (modern != FALSE);
	a->bitmatrix = B_make(x,y);
	/* a->rownonempty = B_make(x,1); */
	/* a->columnnonempty = B_make(1,y); */
	/* If the board has an edge, we cannot expand the board size or
	   shrink it (lest we expand it too much later.) */
	a->north_edge = FALSE;
	a->south_edge = FALSE;
	a->east_edge = FALSE;
	a->west_edge = FALSE;
	enable_interrupts;
	return a;
}

/* This has been inlined, for efficiency; perhaps disable interrupts
   around its occurrences. */
/* Doesn't initialize any of the slots. */
/*
 *  K_type K_make_empty ()
 *  {
 *  	K_type a;
 *
 *  	disable_interrupts;
 *  	if (!(a = (K_type) malloc (sizeof (struct K_struct))))
 *  		myerror ("K_make_empty: malloc couldn't free block");
 *  	enable_interrupts;
 *  	return a;
 *  }
 */

void K_printf (K_type a)
{
  int i,j;

  if (a->north_edge)
    { printf(" ");
      if (a->west_edge) printf("+");
      for (j = 0; j < K_y(a); j++)
	printf("-");
      if (a->east_edge) printf("+");
      printf("\n"); }
  for (i = 0; i < K_x(a); i++)
    { if (i>0) printf("\n");
      printf(" ");
      if (a->west_edge) printf("|");
      for (j = 0; j < K_y(a); j++)
	printf ("%d", K_get (a,i,j) ? 1 : 0);
      if (a->east_edge) printf("|");
      /* printf ("  %d", B_get(K_rownonempty(a), i, 0)); */
    }
  printf("   ");
  /* Test correctness of columnnonempty bits. */
  /* for (j = 0; j < K_y(a); j++)
    printf ("%d", B_get (K_columnnonempty(a), 0, j) ? 1 : 0); */
  printf("   %s, polarity %d\n",
	 K_modern(a) ? "modern" : "ancient", K_polarity(a));
  if (a->south_edge)
    { printf(" ");
      if (a->west_edge) printf("+");
      for (j = 0; j < K_y(a); j++)
	printf("-");
      if (a->east_edge) printf("+");
      printf("\n"); }
}

list_type K_to_list (K_type a)
{
  list_type l;

  l =  B_to_list(K_bitmatrix(a));
  list_prepend(l, K_modern(a)
	       + 2*(a->north_edge) + 4*(a->south_edge)
	       + 8*(a->east_edge) + 16*(a->west_edge));
  return l;
}

/* Essentially the purpose of K_pack is performed by the separation routines,
   so it is no longer used. */
/*
 *  void K_pack (K_type a)
 *  {
 *    int xmin, ymin;
 *
 *    xmin = B_min_occupied_x(K_bitmatrix(a));
 *    ymin = B_max_occupied_x(K_bitmatrix(a));
 *
 *    // Could check for one of these values = -1, but that will never happen
 *    // in Konane.
 *
 *    kdebugprintf(("xmin:%d ymin:%d polarity:%d &1:%d\n",
 *  	       xmin, ymin, K_polarity(a),
 *  	       ((xmin ^ ymin ^ K_polarity(a)) & 1)));
 *
 *    a->polarity = ((xmin ^ ymin ^ K_polarity(a)) & 1);
 *    a->bitmatrix = B_pack(K_bitmatrix(a));
 *    // a->rownonempty = B_pack(K_rownonempty(a));
 *    // a->columnnonempty = B_pack(K_columnnonempty(a));
 *  }
 */

K_type K_copy (K_type a)
{
  K_type b;

  /* Was: b = K_make_empty(); */
  if (!(b = (K_type) malloc (sizeof (struct K_struct))))
    myerror ("K_copy: malloc couldn't free block");

  b->polarity = a->polarity;
  b->modern = a->modern;
  b->bitmatrix = B_copy(a->bitmatrix);
  b->north_edge = a->north_edge;
  b->south_edge = a->south_edge;
  b->east_edge = a->east_edge;
  b->west_edge = a->west_edge;

  /* b->rownonempty = B_copy(a->rownonempty); */
  /* b->columnnonempty = B_copy(a->columnnonempty); */

  return b;
}

/* Returns a K_struct with garbage {north,south,east,west}edge slots. */
K_type K_copy_offset (K_type old, int xoffset, int yoffset, int xsize, int ysize)
{
  K_type New;

  /* Was: New = K_make_empty(); */
  if (!(New = (K_type) malloc (sizeof (struct K_struct))))
    myerror ("K_copy_offset: malloc couldn't free block");

  New->polarity = ((xoffset ^ yoffset ^ K_polarity(old)) & 1);
  New->modern = K_modern(old);
  New->bitmatrix = B_copy_offset(old->bitmatrix, xoffset, yoffset, xsize, ysize);

  return New;
}


/* Does not modify or free its argument; just returns an expanded copy. */
K_type K_expand (K_type a, int dx, int dy)
{
  K_type b;

  /* Was: b = K_make_empty(); */
  if (!(b = (K_type) malloc (sizeof (struct K_struct))))
    myerror ("K_make: malloc couldn't free block");

#if (KDEBUG && EXPAND_DEBUG)
  kdepthkprintf(("Before expansion (%d,%d):\n",dx,dy), a);
#endif

  if (dx + dy > 0)
    /* Added a bottom row or right column. */
    b->polarity = a->polarity;
  else
    /* Added a top row or left column. */
    b->polarity = !(a->polarity);
  b->modern = a->modern;
  b->bitmatrix = B_expand(a->bitmatrix, dx, dy);
  b->north_edge = a->north_edge;
  b->south_edge = a->south_edge;
  b->east_edge = a->east_edge;
  b->west_edge = a->west_edge;

#if (KDEBUG && EXPAND_DEBUG)
  kshowdepth();
  printf("Expanded bitmatrix:\n");
  B_printf(b->bitmatrix);
#endif

  /*  if (dx==0)
   *    { b->rownonempty = B_copy(a->rownonempty);
   *      b->columnnonempty = B_expand(a->columnnonempty, dx, dy); }
   *  else
   *    { b->rownonempty = B_expand(a->rownonempty, dx, dy);
   *      b->columnnonempty = B_copy(a->columnnonempty); }
   */
#if (KDEBUG && EXPAND_DEBUG)
  kdepthkprintf(("After expansion:\n"), b);
#endif

  return b;
}


/* Makes sure a's x dimension is no larger than its y dimension, and
   possibly flips to canonical mirror reversal as well. */
/* Returns the value returned by B_flip. */
int K_orient (K_type a)
{
  int flip_value;
  boolean temp;

  if (K_x(a) > K_y(a))
    { a->bitmatrix = B_transpose(K_bitmatrix(a));
      temp = a->west_edge;
      a->west_edge = a->north_edge;
      a->north_edge = temp;
      temp = a->east_edge;
      a->east_edge = a->south_edge;
      a->south_edge = temp;
      /* temp = B_transpose(a->columnnonempty); */
      /* a->columnnonempty = B_transpose(a->rownonempty); */
      /* Does temp need to be freed?  I think not, as B_transpose frees. */
      /* a->rownonempty = temp; */
    }

  flip_value = B_flip(K_bitmatrix(a));
  /* If 1s bit is set, bitarray was flipped horizontally. */
  if (flip_value & 1)
    { temp = a->west_edge;
      a->west_edge = a->east_edge;
      a->east_edge = temp;
      /* If even Y dimension (# of columns), flip polarity. */
      if (!(K_y(a) & 1))
	K_polarity(a) = !K_polarity(a); }
  /* If 2s bit is set, bitarray was flipped vertically. */
  if (flip_value & 2)
    { temp = a->north_edge;
      a->north_edge = a->south_edge;
      a->south_edge = temp;
      /* If even X dimension (# of rows), flip polarity. */
      if (!(K_x(a) & 1))
	K_polarity(a) = !K_polarity(a); }

  return flip_value;
}

/* Same as game_print, but returns its argument
 * so that it can be passed to list_map. */
game_type K_game_print (game_type g) {
	game_print (g);
	return g;
}


/* ***************************************************************************
   Konane game code
   **************** */


/* Forward declarations. */
game_type jump_konane_single (K_type, int, int, int, int);
game_type jump_konane_multiple_ns (K_type, int, int, int, int);
game_type jump_konane_multiple_ew (K_type, int, int, int, int);
game_type konane_position_nonseparable_modern (K_type pos);
game_type konane_position_nonseparable_ancient (K_type pos);

/* Maintain efficiency constraints on games when they're created, not later.
   That is, in jump_konane and read_konane. */
/* Another possibility:  do so as first action of konane_position. */


/* It used to be the case that these functions shared more code, but
   something was going amiss with allocation and deallocation.  The current
   scheme works, so don't mess with it.
 */

/* This is called by konane_position. */
/* The argument should already be packed and nonseparable. */
/* Does not modify its argument. */
#define konane_position_nonseparable(pos)	\
  (K_modern(pos) ?				\
   konane_position_nonseparable_modern(pos) :	\
   konane_position_nonseparable_ancient(pos))

/* This is called by konane_position. */
/* The argument should already be packed and nonseparable. */
/* Does not modify its argument. */
game_type konane_position_nonseparable_modern (K_type pos)
{
  game_type g;
  int i, j, xmax, ymax;
  list_type left, right, l, this_follower;
  K_type temp_pos;

#if (KDEBUG)
  kdepth++;
#endif

  l = K_to_list (pos);
  if (hash_test (KONANE_KEY, l))
    { list_free(l);
      g = hash_get_last();
      if (K_polarity(pos))
	g = negative(g);
#if (KDEBUG)
      kshowdepth();
      printf("Hashtable result = ");
      game_printf(g);
      kdepth--;
#endif
      return g;
    }
  left = list_make();
  right = list_make();
  xmax = K_x(pos)-1;
  ymax = K_y(pos)-1;
  for (i = 0; i <= xmax; i++)
    for (j = 0; j <= ymax; j++)
      if (K_get (pos, i, j))
	/* check in all four directions */

	/* It might be nice to merge the conditional bodies here.  Perhaps
	   have "add a row" ONLY add a row, and take the other out of the
	   else.  If I'd just added a row, I wouldn't need to check the bit,
	   but I could anyway, to simplify the code.
	   */
	{
	  /* Determines whether the move was made by black or white. */
	  this_follower = ((i ^ j ^ K_polarity(pos)) & 1) ? right : left;

	  /* Should jump_konane figure out whether to add a row?
	     I must remember (or test) whether I've added a row,
	     always add one (which could be very expensive, because
	     of cleanup). */

	  /* Possible cleverness:  keep the expanded matrix around (instead
	     of freeing it) in case later I do a similar expansion -- then
	     it can be reused. */

#if (KDEBUG)
	  kdepthprintf(("M: Found a piece at %d,%d.\n", i, j));
#endif
	  if ((i < xmax) && (K_get (pos, i+1, j)))
	    { /* empty line to make else indent properly */
	      if (i == xmax - 1)
		{ if (!(pos->south_edge))
		    /* add a row on the bottom */
		    { temp_pos = K_expand(pos,1,0);
		      list_insert (this_follower,
				   jump_konane_single(temp_pos,i,j,1,0));
		      K_free(temp_pos); } }
	      else if (!K_get(pos, i+2, j))
		list_insert (this_follower, jump_konane_single(pos,i,j,1,0)); }

	  if ((j < ymax) && (K_get (pos, i, j+1)))
	    {
	      if (j == ymax - 1)
		{ if (!(pos->east_edge))
		    /* add a row on the right */
		    { temp_pos = K_expand(pos,0,1);
		      list_insert (this_follower,
				   jump_konane_single(temp_pos,i,j,0,1));
		      K_free(temp_pos); } }
	      else if (!K_get(pos, i, j+2))
		list_insert (this_follower, jump_konane_single(pos,i,j,0,1)); }

	  if ((i > 0) && (K_get (pos, i-1, j)))
	    {
	      if (i == 1)
		{ if (!(pos->north_edge))
		    /* add a row on the top */
		    { temp_pos = K_expand(pos,-1,0);
		      list_insert (this_follower,
				   jump_konane_single(temp_pos,i+1,j,-1,0));
		      K_free(temp_pos); } }
	      else if (!K_get(pos, i-2, j))
		list_insert (this_follower, jump_konane_single(pos,i,j,-1,0)); }

	  if ((j > 0) && (K_get (pos, i, j-1)))
	    {
	      if (j == 1)
		{ if (!(pos->west_edge))
		    /* add a row on the left */
		    { temp_pos = K_expand(pos,0,-1);
		      list_insert (this_follower,
				   jump_konane_single(temp_pos,i,j+1,0,-1));
		      K_free(temp_pos); } }
	      else if (!K_get(pos, i, j-2))
		list_insert (this_follower, jump_konane_single(pos,i,j,0,-1)); }
	}

  /* I have the left and right followers; merge them. */
#if (KDEBUG)
  kshowdepth();
  printf("Left followers: ");
  list_map(K_game_print, left);
  printf("\n");
  kshowdepth();
  printf("Right followers: ");
  list_map(K_game_print, right);
  printf("\n");
#endif
  /* I believe that because of this use, I needn't list_free left or right. */
  g = make(left, right);
#if (KDEBUG)
  kshowdepth();
  printf("Result = ");
  game_printf(g);
  kdepth--;
#endif
  if (K_polarity(pos))
    /* I don't know why I need the brackets, but I get errors otherwise. */
    { hash_put (KONANE_KEY, l, negative(g)); }
  else hash_put (KONANE_KEY, l, g);
  return g;
}

/* This is called by konane_position. */
/* The argument should already be packed and nonseparable. */
/* Does not modify its argument. */
game_type konane_position_nonseparable_ancient (K_type pos)
{
  game_type g;
  int i, j, xmax, ymax, tmpi, tmpj;
  list_type left, right, l, this_follower;
  K_type temp_pos;

#if (KDEBUG)
  kdepth++;
#endif

  l = K_to_list (pos);
  if (hash_test (KONANE_KEY, l))
    { list_free(l);
      g = hash_get_last();
      if (K_polarity(pos))
	g = negative(g);
#if (KDEBUG)
      kshowdepth();
      printf("Hashtable result = ");
      game_printf(g);
      kdepth--;
#endif
      return g;
    }
  left = list_make();
  right = list_make();
  xmax = K_x(pos)-1;
  ymax = K_y(pos)-1;
  for (i = 0; i <= xmax; i++)
    for (j = 0; j <= ymax; j++)
      if (K_get (pos, i, j))
	/* check in all four directions */

	/* It might be nice to merge the conditional bodies here.  Perhaps
	   have "add a row" ONLY add a row, and take the other out of the
	   else.  If I'd just added a row, I wouldn't need to check the bit,
	   but I could anyway, to simplify the code.
	*/
	{
	  /* Determines whether the move was made by black or white. */
	  this_follower = ((i ^ j ^ K_polarity(pos)) & 1) ? right : left;

	  /* Should jump_konane figure out whether to add a row?
	     I must remember (or test) whether I've added a row,
	     always add one (which could be very expensive, because
	     of cleanup). */

	  /* Possible cleverness:  keep the expanded matrix around (instead
	     of freeing it) in case later I do a similar expansion -- then
	     it can be reused. */

#if (KDEBUG)
	  kdepthprintf(("A: Found a piece at %d,%d.\n", i, j));
#endif

	  tmpi = i;
	  while ((tmpi < xmax) && (K_get (pos, tmpi+1, j)))
	    { /* empty line to make else indent properly */
#if (KDEBUG)
	      kdepthprintf(("Candidate jumpee at %d,%d.\n", tmpi+1, j));
#endif
	      if (tmpi == xmax - 1)
		{ if (!(pos->south_edge))
		    /* add a row on the bottom */
		    { temp_pos = K_expand(pos,1,0);
		      list_insert (this_follower,
				   jump_konane_multiple_ns(temp_pos,i,j,1,tmpi+1));
		      K_free(temp_pos); } }
	      else if (!K_get(pos, tmpi+2, j))
		{ list_insert (this_follower,
			       jump_konane_multiple_ns(pos,i,j,1,tmpi+1));
		  tmpi = tmpi + 2;
		  continue; }
	      break;
	    }

	  tmpj = j;
	  while ((tmpj < ymax) && (K_get (pos, i, tmpj+1)))
	    {
#if (KDEBUG)
	      kdepthprintf(("Candidate jumpee at %d,%d.\n", i, tmpj+1));
#endif
	      if (tmpj == ymax - 1)
		{ if (!(pos->east_edge))
		    /* add a row on the right */
		    { temp_pos = K_expand(pos,0,1);
		      list_insert (this_follower,
				   jump_konane_multiple_ew(temp_pos,i,j,1,tmpj+1));
		      K_free(temp_pos); } }
	      else if (!K_get(pos, i, tmpj+2))
		{ list_insert (this_follower,
			       jump_konane_multiple_ew(pos,i,j,1,tmpj+1));
		  tmpj = tmpj + 2;
		  continue; }
	      break;
	    }

	  tmpi = i;
	  while ((tmpi > 0) && (K_get (pos, tmpi-1, j)))
	    {
#if (KDEBUG)
	      kdepthprintf(("Candidate jumpee at %d,%d.\n", tmpi-1, j));
#endif
	      if (tmpi == 1)
		{ if (!(pos->north_edge))
		    /* add a row on the top */
		    { temp_pos = K_expand(pos,-1,0);
		      /* i+1 instead of i (and tmpi instead of tmpi-1) because
			 addition of a row shifted all the contents right */
		      list_insert (this_follower,
				   jump_konane_multiple_ns(temp_pos,i+1,j,-1,tmpi));
		      K_free(temp_pos); } }
	      else if (!K_get(pos, tmpi-2, j))
		{ list_insert (this_follower,
			       jump_konane_multiple_ns(pos,i,j,-1,tmpi-1));
		  tmpi = tmpi - 2;
		  continue; }
	      break;
	    }

	  tmpj = j;
	  while ((tmpj > 0) && (K_get (pos, i, tmpj-1)))
	    {
#if (KDEBUG)
	      kdepthprintf(("Candidate jumpee at %d,%d.\n", i, tmpj-1));
#endif
	      if (tmpj == 1)
		{ if (!(pos->west_edge))
		    /* add a row on the left */
		    { temp_pos = K_expand(pos,0,-1);
		      /* j+1 instead of j (and tmpj instead of tmpj-1) because
			 addition of a row shifted all the contents */
		      list_insert (this_follower,
				   jump_konane_multiple_ew(temp_pos,i,j+1,-1,tmpj));
		      K_free(temp_pos); } }
	      else if (!K_get(pos, i, tmpj-2))
		{ list_insert (this_follower,
			       jump_konane_multiple_ew(pos,i,j,-1,tmpj-1));
		  tmpj = tmpj - 2;
		  continue; }
	      break;
	    }
	}

  /* I have the left and right followers; merge them. */
#if (KDEBUG)
  kshowdepth();
  printf("Left followers: ");
  list_map(K_game_print, left);
  printf("\n");
  kshowdepth();
  printf("Right followers: ");
  list_map(K_game_print, right);
  printf("\n");
#endif
  /* I believe that because of this use, I needn't list_free left or right. */
  g = make(left, right);
#if (KDEBUG)
  kshowdepth();
  printf("Result = ");
  game_printf(g);
  kdepth--;
#endif
  if (K_polarity(pos))
    /* I don't know why I need the brackets, but I get errors otherwise. */
    { hash_put (KONANE_KEY, l, negative(g)); }
  else hash_put (KONANE_KEY, l, g);
  return g;
}


/* One of dx, dy is 0; the other is +/- 1. */
/* This function assumes that x,y has a stone, as does x+dx,y+dy, but
   x+2*dx,y+2*dy is empty. */
/* Does not modify its argument; the bitmatrix should already be big enough. */
/* Returns the value of the resulting game. */
game_type jump_konane_single (K_type pos, int x, int y, int dx, int dy)
{
  game_type g;
  /* boolean jumpernonempty = FALSE; */
  /* boolean jumpeenonempty = FALSE; */
  K_type resultpos;
  /* int i, j; */

  resultpos = K_copy(pos);

  K_clear(resultpos, x, y);
  K_clear(resultpos, x+dx, y+dy);
  B_set(K_bitmatrix(resultpos), x+dx+dx, y+dy+dy);

#if (KDEBUG)
  kdepthkprintf(("Jumped from %d,%d to %d,%d, resulting in:\n",
		 x, y, x+dx+dx, y+dy+dy), resultpos);
#endif

#if K1D_OPTIMIZE
  if (dy && (K_x(resultpos) == 1))
    { g = konane_position_1d(resultpos, y, dy); }
  else
#endif  /* K1D_OPTIMIZE */
    {
      /* K_pack(resultpos); */
      K_orient(resultpos);

      g = konane_position(resultpos);
    }

  K_free(resultpos);
  return(g);
}


/* One of dx, dy is 0; the other is +/- 1. */
/* This function assumes that x,y has a stone, as does x+dx,y+dy, but
   x+2*dx,y+2*dy is empty. */
/* Does not modify its argument; the bitmatrix should already be big enough. */
/* Returns the value of the resulting game. */
game_type jump_konane_multiple_ns (K_type pos, int x, int y, int dx, int limit)
{
  game_type g;
  K_type resultpos;
  int i;

#if (KDEBUG)
  kdepthkprintf(("jump_konane_multiple_ns %d %d %d %d\n", x, y, dx, limit),pos);
#endif

  resultpos = K_copy(pos);

  K_clear(resultpos, x, y);
  for (i = x + dx; i != limit + 2*dx; i += 2*dx)
    K_clear(resultpos, i, y);
  K_set(resultpos, limit+dx, y);

#if (KDEBUG)
  kdepthkprintf(("Jumped from %d,%d to %d,%d, resulting in:\n",
		 x, y, limit+dx, y), resultpos);
#endif

  /* No 1d optimization here; should there be? */

  /* K_pack(resultpos); */
  K_orient(resultpos);

  g = konane_position(resultpos);

  K_free(resultpos);
  return(g);
}


/* One of dx, dy is 0; the other is +/- 1. */
/* This function assumes that x,y has a stone, as does x+dx,y+dy, but
   x+2*dx,y+2*dy is empty. */
/* Does not modify its argument; the bitmatrix should already be big enough. */
/* Returns the value of the resulting game. */
game_type jump_konane_multiple_ew (K_type pos, int x, int y, int dy, int limit)
{
  game_type g;
  K_type resultpos;
  int j;

#if (KDEBUG)
  kdepthkprintf(("jump_konane_multiple_ew %d %d %d %d\n", x, y, dy, limit),pos);
#endif

  resultpos = K_copy(pos);

  K_clear(resultpos, x, y);
  for (j = y + dy; j != limit + 2*dy; j += 2*dy)
    K_clear(resultpos, x, j);
  K_set(resultpos, x, limit+dy);

#if (KDEBUG)
  kdepthkprintf(("Jumped from %d,%d to %d,%d, resulting in:\n",
		 x, y, x, limit+dy), resultpos);
#endif

#if K1D_OPTIMIZE
  if (K_x(resultpos) == 1)
    { g = konane_position_1d(resultpos, y, dy); }
  else
#endif  /* K1D_OPTIMIZE */
    {
      /* K_pack(resultpos); */
      K_orient(resultpos);

      g = konane_position(resultpos);
    }

  K_free(resultpos);
  return(g);
}


/* Input constraints:
 *   Use - on the first row to specify north edge.  You may not leave a blank
 * row before a row with -, lest it be considered the south edge!
 *   Use - on the last row to specify south edge.
 *   Don't use - otherwise.
 *   Use | in the first column to specify west edge.  You may not leave a blank
 * column before a column with |, lest it be considered the east edge!
 *   Use | in the last column to specify east edge.
 *   Don't use | otherwise.
 *   You may use + in any of the corners, if you like.
 *   Don't use it anywhere else.
 *   Otherwise, any character except a space counts as a stone.
 *   The upper left-hand corner is a black square.
 *   No checking for X=black, O=white is done.
 *  I make no guarantees about what happens if these constraints are not followed.
 */

K_type input_konane(int modern)
{
  K_type pos;
  char s[KONANE_INPUT_HEIGHT][MAXBUFF];
  int rows, maxcolumn, i, j;
  B_type bitmatrix;
  char thischar;

  boolean north_face=FALSE, west_face=FALSE, south_face=FALSE, east_face=FALSE;
  boolean stone_in_left_column=FALSE, stone_in_this_row=FALSE;

  bitmatrix = B_make(KONANE_INPUT_HEIGHT,KONANE_INPUT_WIDTH);
  printf ("Enter %s Konane position followed by extra <cr>\n",
	  modern ? "modern" : "ancient");
  maxcolumn = 0;
  for (rows=0;
       input_line(s[rows]) != NULL && *(s[rows]) != '\0' && rows < KONANE_INPUT_HEIGHT;
       rows++)
  /* Compute size of needed array as we read the character representation. */
    { stone_in_this_row = FALSE;
      for (j=0; (thischar = *(s[rows]+j)) != '\0'; j++)
	{ /* Left or right board edge. */
	  if ((thischar == '|') || (thischar == '+'))
	    {
	      if (j)
		{ if (maxcolumn > j-1)
		    printf("Warning!  There are stones off the right board edge.\n");
		  maxcolumn = j-1;
		  east_face = TRUE; }
	      else
		{ if (stone_in_left_column)
		    printf("Warning!  There are stones off the left board edge.\n");
		  west_face=TRUE; } }
	  /* Top or bottom board edge. */
	  if ((thischar == '-') || (thischar == '+'))
	    {
	      if (rows)
		{ if (stone_in_this_row)
		    printf("Warning!  There are stones off the bottom board edge.\n");
		  south_face = TRUE; }
	      else
		{ if (stone_in_this_row)
		    printf("Warning!  There are stones off the top board edge.\n");
		  north_face=TRUE; } }
	  /* Stone. */
	  if ((thischar != ' ') && (thischar != '+')
	      && (thischar != '|') && (thischar != '-'))
	    { if (j)
	      { if (j > maxcolumn)
		  { if (east_face)
		      printf("Warning!  There are stones off the right board edge.\n");
		    else
		      maxcolumn = j; } }
	    else
	      { if (west_face)
		  printf("Warning!  There are stones off the left board edge.\n");
		else
		  stone_in_left_column = TRUE; }

	    stone_in_this_row = TRUE;
	    if (rows)
	      { if (south_face)
		  printf("Warning!  There are stones off the bottom board edge.\n"); }
	    else
	      if (north_face)
		printf("Warning!  There are stones off the top board edge.\n"); }
      } } /* end of for loops */

  if (rows == KONANE_INPUT_HEIGHT)
    printf("Warning!  There may only be %d rows of stones.\n",
	   KONANE_INPUT_HEIGHT);

#if (KDEBUG)
  kdepthprintf(("Input of stones; rows=%d, maxcolumn=%d, faces=%d %d %d %d\n",
		rows, maxcolumn, north_face, south_face, west_face, east_face));
#endif

  /* I could have a check here for rows and maxcolumns both == 0 (in which
     case just return zero immediately). */

  rows = rows - north_face - south_face;
  /* east_face has already been dealt with */
  maxcolumn = maxcolumn - west_face;

#if (KDEBUG)
  kdepthprintf(("Modified input of stones; rows=%d, maxcolumn=%d, faces=%d %d %d %d\n",
		rows, maxcolumn, north_face, south_face, west_face, east_face));
#endif

  pos = K_make(rows, maxcolumn+1, modern);
  pos->north_edge = north_face;
  pos->south_edge = south_face;
  pos->west_edge = west_face;
  pos->east_edge = east_face;
  /* pos->rownonempty = B_make(B_x(bitmatrix), 1); */
  /* pos->columnnonempty = B_make(1, B_y(bitmatrix)); */

  if (maxcolumn >= KONANE_INPUT_WIDTH)
    { printf("Warning!  Input lines truncated to width of %d characters.\n",
	     KONANE_INPUT_WIDTH);
      maxcolumn = KONANE_INPUT_WIDTH - 1;
    }

  for (i=0; i<rows; i++)
  /* Compute size of needed array as we read the character representation. */
    for (j=0;
	 (j<=maxcolumn) && (thischar = *(s[i+north_face]+j+west_face)) != '\0';
	 j++)
      if (thischar != ' ') K_set (pos, i, j);

  /*  for (i=0; i<B_x(bitmatrix); i++)
   *    for (j=0; j<B_y(bitmatrix); j++)
   *      if (B_get(bitmatrix, i, j))
   *	{ B_set(K_rownonempty(pos), i, 0);
   *	  B_set(K_columnnonempty(pos), 0, j); }
   */

  /* Do this after setting polarity=0 in order to maintain polarity. */

  /* kdebugkprintf(("Before packing:\n"), pos); */
  /* K_pack(pos); */
  /* kdebugkprintf(("After packing:\n"), pos); */

#if (KDEBUG)
  kdepthkprintf(("Before orienting:\n"), pos);
#endif
  K_orient(pos);
#if (KDEBUG)
  kdepthkprintf(("After orienting:\n"), pos);
#endif

  return pos;
}

/* ignore the list_type argument for now */
game_type konane (list_type l, int modern)
{
  K_type pos;
  game_type g;

  pos = input_konane(modern);
  kdepth = 0;
#if (KDEBUG)
  kdepthkprintf(("Konane position:\n"), pos);
#endif
  K_orient(pos);
#if (KDEBUG)
  kdepthkprintf(("Oriented Konane position:\n"), pos);
#endif
  g = konane_position(pos);
  K_free(pos);
  return g;
}


/* ***************************************************************************
   Separating
   ********** */


/* arguments are depth, width */
int penarray_by_size[6][14] =
/* Must spell out contents of first row for some compilers. */
{ { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
  { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
  { 0, 1, 1, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3 },
  { 0, 1, 1, 2, 2, 3, 3, 3, 4, 4, 4, 4, 4, 4 },
  { 0, 1, 1, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4 },
  { 0, 1, 1, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4 } };

#define penetration_by_size(depth, width) \
  (penarray_by_size[min(depth,5)][min(width,13)])

int penarray_by_stones[20] =
{ 0, 0, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3};

/* Not indexing with a min() expression means marginally better performance
   in the larger case. */
#define penetration_by_stones(stones) \
  ((stones > 19) ? 4 : penarray_by_stones[stones])


/* This function simply separates the independent components, then calls
   konane_position_nonseparable to evaluate them. */
/* It looks like this is a lot of work, but it really does save time. */
/* The argument should already be packed and oriented, but may be separable. */
/* Does not modify its argument. */
game_type konane_position (K_type pos)
{
  game_type g;
  K_type New;
  B_type stones_unturned;
  int i, j, k;
  boolean expanded;
  int xmax, ymax, xsize, ysize;
  int pen_limit, pen_vert, pen_horiz;

  /* The values of the array moat variables are equal to penetration.
     Values of moat scalars are one greater than the penetration.  When
     comparing an array vs a scalar, check for overlap, not adjacency.
     This is effectively a test for adjacency in the penetrated-to regions.
     Need to recompute moat values when put in the array (I presume I do).
     */

  /* These are declared inside the function because of recursive calls. */

  /* Should I instead have a struct of 8 values and a single array?
     Or a single array of higher dimension? */
  int groups = 0;
  int stones[KONANE_MAX_GROUPS];
  int stonexmin[KONANE_MAX_GROUPS], stonexmax[KONANE_MAX_GROUPS];
  int stoneymin[KONANE_MAX_GROUPS], stoneymax[KONANE_MAX_GROUPS];
  int moatxmin[KONANE_MAX_GROUPS], moatxmax[KONANE_MAX_GROUPS];
  int moatymin[KONANE_MAX_GROUPS], moatymax[KONANE_MAX_GROUPS];
  int thisstones;
  int thisstonexmin, thisstonexmax, thisstoneymin, thisstoneymax;
  int thismoatxmin, thismoatxmax, thismoatymin, thismoatymax;
  /* int thischeckedxmin; */
  int thischeckedxmax, thischeckedymin, thischeckedymax;
  /* Actually, because of the direction of search, I don't need
     thischeckedxmin because I will never need to check in the upward
     direction. */

  g = zero;

  /* The goal is to leave no stone unturned. */
  stones_unturned = B_copy(pos->bitmatrix);

  /* When I find a stone, create a new group and augment its size as much as
     possible; then add the group to the list of groups. */

  /* Should I subtract one from these values?  Probably. */
  xmax = B_x(stones_unturned)-1;
  ymax = B_y(stones_unturned)-1;
  for (i = 0; i <= xmax; i++)
    for (j = 0; j <= ymax; j++)
      if (B_get (stones_unturned, i, j))
	{
#if (KDEBUG)
	  kdepthprintf(("Found a group at (%d, %d).\n", i, j));
#endif

	  B_clear (stones_unturned, i, j);
	  thisstones = 1;

	  thisstonexmin = i; thisstonexmax = i;
	  thisstoneymin = j; thisstoneymax = j;
	  thischeckedxmax = i;
	  thischeckedymin = j; thischeckedymax = j;

	  expanded = TRUE;
	  while (expanded)
	    { expanded = FALSE;

	      xsize = thisstonexmax-thisstonexmin+1;
	      ysize = thisstoneymax-thisstoneymin+1;

	      /* For vertical penetration, the number of rows is the depth
		 and the number of columns is the width.  And vice-versa. */
	      pen_limit = penetration_by_stones(thisstones);
	      pen_vert = penetration_by_size(xsize, ysize);
	      pen_vert = min(pen_limit, pen_vert);
	      pen_horiz = penetration_by_size(ysize, xsize);
	      pen_horiz = min(pen_limit, pen_horiz);

	      thismoatxmin = max(0, thisstonexmin - pen_vert - 1);
	      thismoatxmax = min(xmax, thisstonexmax + pen_vert + 1);
	      thismoatymin = max(0, thisstoneymin - pen_horiz - 1);
	      thismoatymax = min(ymax, thisstoneymax + pen_horiz + 1);

	      while (thischeckedxmax < thismoatxmax)
		{ thischeckedxmax++;
		  for (k=thischeckedymin; k<=thischeckedymax; k++)
		    if (B_get(stones_unturned, thischeckedxmax, k))
		      { thisstonexmax = thischeckedxmax;
			/* I wish I didn't have to do this, but I think I do. */
			thisstoneymin = min(thisstoneymin, k);
			thisstoneymax = max(thisstoneymax, k);
			B_clear(stones_unturned, thischeckedxmax, k);
			thisstones++;
			expanded = TRUE; } }
	      while (thischeckedymin > thismoatymin)
		{ thischeckedymin--;
		  for (k=i+1; k<=thischeckedxmax; k++)
		    if (B_get(stones_unturned, k, thischeckedymin))
		      { thisstoneymin = thischeckedymin;
			thisstonexmin = min(thisstonexmin, k);
			thisstonexmax = max(thisstonexmax, k);
			B_clear(stones_unturned, k, thischeckedymin);
			thisstones++;
			expanded = TRUE; } }
	      while (thischeckedymax < thismoatymax)
		{ thischeckedymax++;
		  for (k=i; k<=thischeckedxmax; k++)
		    if (B_get(stones_unturned, k, thischeckedymax))
		      { thisstoneymax = thischeckedymax;
			thisstonexmin = min(thisstonexmin, k);
			thisstonexmax = max(thisstonexmax, k);
			B_clear(stones_unturned, k, thischeckedymax);
			thisstones++;
			expanded = TRUE; } }
	      /* kdebugprintf(("Done with moat checking; expanded = %d.\n", expanded)); */

	      /* Hold off on this check until we've done as much of the
		 other one as possible. */
	      if (!expanded)
		/* Check for overlap with groups in the arrays. */
		for (k=0; k<groups; k++)
		  if ((thismoatxmax >= moatxmin[k])
		      && (thismoatxmin <= moatxmax[k])
		      && (thismoatymax >= moatymin[k])
		      && (thismoatymin <= moatymax[k]))
		    /* There is an overlap. */
		    { thisstonexmin = min(thisstonexmin, stonexmin[k]);
		      thisstonexmax = max(thisstonexmax, stonexmax[k]);
		      thisstoneymin = min(thisstoneymin, stoneymin[k]);
		      thisstoneymax = max(thisstoneymax, stoneymax[k]);
		      thisstones = thisstones + stones[k];
		      expanded = TRUE;
		      /* remove the array element */
		      groups--;
		      if (groups)
			{ stonexmin[k] = stonexmin[groups];
			  stonexmax[k] = stonexmax[groups];
			  stoneymin[k] = stoneymin[groups];
			  stoneymax[k] = stoneymax[groups];
			  moatxmin[k] = moatxmin[groups];
			  moatxmax[k] = moatxmax[groups];
			  moatymin[k] = moatymin[groups];
			  moatymax[k] = moatymax[groups];
			  /* get out of this loop */
			  /* once I've found one overlap, don't look for more */
			  k = groups; }
		    }

	    }  /* end of while (expanded) */

	  /* Add this group to the array. */
	  /* Stone values are valid. */
	  stones[groups] = thisstones;
	  stonexmin[groups] = thisstonexmin;
	  stonexmax[groups] = thisstonexmax;
	  stoneymin[groups] = thisstoneymin;
	  stoneymax[groups] = thisstoneymax;
	  /* Moat values should be exact penetration, not 1+penetration,
	     and not stopping at board edges. */
	  moatxmin[groups] = thisstonexmin - pen_vert;
	  moatxmax[groups] = thisstonexmax + pen_vert;
	  moatymin[groups] = thisstoneymin - pen_horiz;
	  moatymax[groups] = thisstoneymax + pen_horiz;
#if (KDEBUG)
	  kdepthprintf(("Added group to the array: %d-%d %d-%d; moats: %d-%d %d-%d.\n",
			thisstonexmin, thisstonexmax,
			thisstoneymin, thisstoneymax,
			moatxmin[groups], moatxmax[groups],
			moatymin[groups], moatymax[groups]));
#endif
	  groups++;

	} /* end of for loops */

#if (KDEBUG)
  kdepthprintf(("Found a total of %d groups.\n",groups));
#endif
  /* Now all of the groups are independent and may be evaluated separately. */
  /* Don't do this because no packing will get done, even though no new K_type
     will get allocated. */
  /*  if (groups==1)
   *    g = plus(g, konane_position_nonseparable(pos));
   *  else
   */
    for (k=0; k<groups; k++)
      {
	/* Make sure edges are included if necessary. */

	if ((moatxmin[k] < 0) && pos->north_edge)
	  stonexmin[k]=0;
	if ((moatxmax[k] > xmax) && pos->south_edge)
	  stonexmax[k]=xmax;
	if ((moatymin[k] < 0) && pos->west_edge)
	  stonexmin[k]=0;
	if ((moatymax[k] > ymax) && pos->east_edge)
	  stoneymax[k]=ymax;

	New = K_copy_offset(pos, stonexmin[k], stoneymin[k],
			    stonexmax[k]-stonexmin[k]+1,
			    stoneymax[k]-stoneymin[k]+1);

	New->north_edge = ((moatxmin[k] < 0) && pos->north_edge);
	New->south_edge = ((moatxmax[k] > xmax) && pos->south_edge);
	New->west_edge = ((moatymin[k] < 0) && pos->west_edge);
	New->east_edge = ((moatymax[k] > ymax) && pos->east_edge);

	K_orient(New);
#if (KDEBUG)
	kdepthkprintf(("Group %d/%d of %d stones is:\n", k+1, groups, stones[k]),
		      New);
#endif
	g = plus(g, konane_position_nonseparable(New));
	K_free(New); }

  B_free(stones_unturned);

  return g;
}

#if K1D_OPTIMIZE

/* Add a special hash type for 1d konane boards?  Probably not worth it... */

/* Like konane_position, this separates stones and calls
konane_position_nonseparable.  It should only be called when the position
is one-dimensional (K_x(pos) == 1) and a jump from 1,from_y to 1,to_y (over
an intervening square) has just ocurred (ie, to_y = from_y +/- 2).  There
is no need to call K_orient before this procedure; in fact, that could flip
the board, invalidating the y and dy values.

This is much faster than the general separating code, and it's even more
thorough.
It could be cleverer by computing actual penetrations, but is that worth it?
*/

/* What should the arguments be?  (As few as possible, I guess.)  */
game_type konane_position_1d (K_type pos, int y, int dy)
{
  game_type g;

  K_type New;

  int ymax = K_y(pos) - 1;

  int from_y = y;
  int to_y = y+dy+dy;

  int ysize = 0;
  int yoffset = 0;

  boolean keep_west_edge = TRUE;
  boolean keep_east_edge = TRUE;

  boolean edge_trim = FALSE;

  k1ddebugkprintf(("kp1d:  y: %d, dy: %d\n", y, dy), pos);

  /* ysize can be set to 0 when ymax is 2; this happens when we start with
     **, expand by 1 to make room for the jump, and end up with __* or *__.
     This is why I can't use a test for ysize != 0.
     It might be nice to get rid of a variable (say, by reusing ymax; but
     let the compiler deal with that, not me.  I want this to be correct!
     */

  if (from_y == 0)
    {
      /* get rid of leftmost two positions */
      ysize = ymax - 1;
      k1ddebugprintf(("kp1d:  ysize set to %d = %d - 1.\n", ysize, ymax));
      yoffset = 2;
      keep_west_edge = FALSE;
      edge_trim = TRUE;
    }
  else if ((to_y == 0) || (from_y == 1))
    {
      /* get rid of leftmost three positions */
      ysize = ymax - 2;
      k1ddebugprintf(("kp1d:  ysize set to %d = %d - 2.\n", ysize, ymax));
      yoffset = 3;
      while ((ysize > 1) && (! K_get(pos, 0, yoffset)))
	{ ysize--;
	  yoffset++; }
      keep_west_edge = FALSE;
      edge_trim = TRUE;
    }
  else if (from_y == ymax)
    {
      /* get rid of rightmost two positions */
      ysize = ymax - 1;
      k1ddebugprintf(("kp1d:  ysize set to %d = %d - 1.\n", ysize, ymax));
      keep_east_edge = FALSE;
      edge_trim = TRUE;
    }
  else if ((to_y == ymax) || (from_y == ymax - 1))
    {
      /* get rid of rightmost three positions */
      ysize = ymax - 2;
      k1ddebugprintf(("kp1d:  ysize set to %d = %d - 2.\n", ysize, ymax));
      while ((ysize > 1) && (! K_get(pos, 0, ysize-1)))
	{ ysize--; }
      keep_east_edge = FALSE;
      edge_trim = TRUE;
    }

  if (edge_trim)
    /* Trim from one of the ends. */
    { if (ysize > 1)
	{
	  New = K_copy_offset(pos, 0, yoffset, 1, ysize);
	  New->north_edge = FALSE;
	  New->south_edge = FALSE;
	  New->west_edge = (pos->west_edge && keep_west_edge);
	  New->east_edge = (pos->east_edge && keep_east_edge);

	  k1ddebugkprintf(("kp1d:  trimmed.  yoffset= %d, ysize= %d.\n",
			   yoffset, ysize),
			  New);

	  g = konane_position_nonseparable(New);
	  K_free(New);
	}
      else
	{
#if (KDEBUG)
	  kdepthprintf(("kp1d:  trimmed to ysize = %d and so ignored.\n", ysize));
#endif
	  /* ysize < 2 */
	  g = zero;
	}
    }
  else if (!K_get(pos, 0, y-dy))
    /* Gap of size at least 3, from y-1 to y+1. */
    /* We know by now that from_y was not an edge or adjacent to one. */
    /* I'd never create empty positions if I knew there were no empty
       border positions before the jump; but I probably don't anyway. */
    {
      if (y > 2)
	{ New = K_copy_offset(pos, 0, 0, 1, y-1);
	  New->north_edge = FALSE;
	  New->south_edge = FALSE;
	  New->west_edge = pos->west_edge;
	  New->east_edge = FALSE;
	  k1ddebugkprintf(("kp1d:  gap of size three found around %d.  Left half (offset 0, ysize %d):\n", y, y-1),
			  New);
	  g = konane_position_nonseparable(New);
	  K_free(New); }
      else
	{ g = zero;
	  k1ddebugprintf(("kp1d:  gap of size three found around %d.  Left half (offset 0, ysize %d) ignored.\n", y, y-1));
	}

      if (y < ymax-2)
	{ New = K_copy_offset(pos, 0, y+2, 1, ymax-y-1);
	  New->north_edge = FALSE;
	  New->south_edge = FALSE;
	  New->west_edge = FALSE;
	  New->east_edge = pos->east_edge;
	  k1ddebugkprintf(("kp1d:  gap of size three found around %d.  Right half (offset %d, ysize %d):\n", y, y+2, ymax-y-1),
			  New);
	  g = plus(g, konane_position_nonseparable(New));
	  K_free(New); }
      else
	k1ddebugprintf(("kp1d:  gap of size three found around %d.  Right half (offset 0, ysize %d) ignored.\n", y, ymax-y-1));

    }
  else
    /* No separation could be done. */
    g = konane_position_nonseparable(pos);

#if (KDEBUG && K1D_DEBUG)
  kshowdepth();
  printf("kp1d:  result = ");
  game_printf(g);
#endif

  return g;
}

#endif /* K1D_OPTIMIZE */
