/* konane.h */
/* by Michael Ernst, December 1991.  Last modified 1/20/92. */
/* Copyright (C) Michael Ernst, 1991.  All rights reserved. */

/* Evaluates Konane positions:
   for example, the position
O
X
 XOX
   is worth -1/2.
*/

/* Should I have a bitmatrix of four edge booleans, or keep them separate?
   Issues of: clarity of code, speed of access, space usage, converting to list.
   */
typedef struct K_struct {
  boolean polarity;		/* 1 ==> upper left hand corner = white */
  boolean modern;
  B_type bitmatrix;		/* x by y */
  boolean north_edge;
  boolean south_edge;
  boolean east_edge;
  boolean west_edge;
  /* B_type rownonempty; */		/* x by 1 */
  /* B_type columnnonempty; */		/* 1 by y */
} *K_type;

#define K_polarity(a)	((a)->polarity)
#define K_modern(a)	((a)->modern)
/* #define K_bitarray(a)	((a)->bitmatrix) */
#define K_bitmatrix(a)	((a)->bitmatrix)
/* #define K_rownonempty(a)	((a)->rownonempty) */
/* #define K_columnnonempty(a)	((a)->columnnonempty) */
#define K_x(a) (B_x(K_bitmatrix(a)))
#define K_y(a) (B_y(K_bitmatrix(a)))
#define K_get(a,i,j)	(B_get(K_bitmatrix(a),i,j))
#define K_set(a,i,j)	(B_set(K_bitmatrix(a),i,j))
#define K_clear(a,i,j)	(B_clear(K_bitmatrix(a),i,j))
/* #define K_free(a)	(B_free ((a)->bitmatrix), B_free((a)->rownonempty), B_free((a)->columnnonempty), free(a)) */
#define K_free(a)	(B_free ((a)->bitmatrix), free(a))

/* Crude min and max */
#define min(a, b)	((a) < (b) ? (a) : (b))
#undef max
#define max(a, b)	((a) > (b) ? (a) : (b))

/* Max number of rows for Konane positions read from command line. */
#define KONANE_INPUT_HEIGHT 20
#define KONANE_INPUT_WIDTH 100	/* It's nice if this equals MAXBUFF.  */
/* Maximum number of groups into which a Konane position may decompose. */
#define KONANE_MAX_GROUPS 32
/* Maximum number of stones per group.  Must be >20. */
/* #define KONANE_MAX_STONES_PER_GROUP 200 */

/* Only call these if debugging is turned on. */
#define kdepthprintf(args) \
  if ((KDEBUG_DEPTH_LIMIT < 0) || (kdepth < KDEBUG_DEPTH_LIMIT)) \
    { kshowdepth(); \
      printf args; }
#define kdepthkprintf(args,k) \
  if ((KDEBUG_DEPTH_LIMIT < 0) || (kdepth < KDEBUG_DEPTH_LIMIT)) \
    { kshowdepth(); \
      printf args; \
      K_printf(k); }

#define k1ddebugprintf(args) \
  if (K1D_DEBUG && ((KDEBUG_DEPTH_LIMIT < 0) || (kdepth < KDEBUG_DEPTH_LIMIT))) \
    { kshowdepth(); \
      printf args; }
#define k1ddebugkprintf(args,k) \
  if (K1D_DEBUG && ((KDEBUG_DEPTH_LIMIT < 0) || (kdepth < KDEBUG_DEPTH_LIMIT))) \
    { kshowdepth(); \
      printf args; \
      K_printf(k); }

extern game_type	konane (list_type, int);
extern game_type	konane_position (K_type);
extern game_type	konane_position_1d (K_type, int, int);
extern K_type		input_konane(int);
