/* Copyright (C) David Wolfe, 1990.  All rights reserved. */

#include <stdlib.h>
#include <signal.h>
#include <stdio.h>
#include "externs.h"
#include "list.h"

unsigned long list_count=0;
static list_type list_free_list = NIL;
       list_type list_0 = NIL;
static list_type ll1 = NIL;		/* A list of length 1 */
static list_type ll2 = NIL;		/*           length 2 */
static list_type ll3 = NIL;		/*           length 3 */

static list_type list_get_cell (void) {
	list_type cell, p;
	int i;

	disable_interrupts;
	if (list_free_list == NIL) {
		if (!(list_free_list = (list_type)
		      malloc(LIST_CELL_BLOCK_SIZE * sizeof (struct list_struct))))
			myerror("list_get_cell: malloc couldn't free block");
		for (i = 1, p=list_free_list; i < LIST_CELL_BLOCK_SIZE;i++, p++)
			p->next = p+1;
		p->next = NIL;
	}
	cell = list_free_list;
	list_free_list = list_free_list->next;
	list_count++;
	enable_interrupts;
	return cell;
}

static void list_free_cell (list_type l) {
	disable_interrupts;
	l -> next = list_free_list;
	list_free_list = l;
	list_count--;
	enable_interrupts;
	return;
}

void list_init(void) {
	if (list_0 == NIL) {
		list_0 = list_make();
		ll1 = list_make();
		list_insert (ll1, 0);
		ll2 = list_make();
		list_insert (ll2, 0);
		list_insert (ll2, 0);
		ll3 = list_make();
		list_insert (ll3, 0);
		list_insert (ll3, 0);
		list_insert (ll3, 0);
	}
	return;
}

list_type list_1(list_elt_type e) {
	ll1->next->elt = e;
	return ll1;
}

list_type list_unordered_2(list_elt_type e1, list_elt_type e2) {
	if (list_elt_less (e1, e2)) {
		ll2->next->elt = e1;
		ll2->next->next->elt = e2;
	} else {
		ll2->next->elt = e2;
		ll2->next->next->elt = e1;
	}
	return ll2;
}

list_type list_2(list_elt_type e1, list_elt_type e2) {
	ll2->next->elt = e1;
	ll2->next->next->elt = e2;
	return ll2;
}

list_type list_3(list_elt_type e1, list_elt_type e2, list_elt_type e3) {
	ll3->next->elt = e1;
	ll3->next->next->elt = e2;
	ll3->next->next->next->elt = e3;
	return ll3;
}

list_type list_make(void) {
	list_type header;

	header = list_get_cell();
	header -> next = NIL;
	list_elt_small (header -> elt);
	return header;
}

void list_insert (list_type l, list_elt_type e) {
	list_type cell;

	while (l->next != NIL && (list_elt_less (l->next->elt, e)))
		l = l->next;
	cell = list_get_cell();
	cell->elt = e;
	cell->next = l->next;
	l->next = cell;
	return;
}

void list_prepend (list_type l, list_elt_type e) {
	list_type cell;

	cell = list_get_cell();
	cell->elt = e;
	cell->next = l->next;
	l->next = cell;
	return;
}

boolean list_member (list_type l, list_elt_type e) {
	while ((l = l -> next) != NIL)
		if (list_elt_equal (l->elt, e))
			return TRUE;
	return FALSE;
}

int list_search (list_type l, list_elt_type e) {
	int i = 0;

	while ((l = l->next) != NIL) {
		i++;
		if (list_elt_equal (l->elt, e))
			return i;
	}
	return 0;
}

int list_length (list_type l) {
	int i = 0;
	
	while ((l = l->next) != NIL)
		i++;
	return i;
}

void list_delete (list_type l, list_elt_type e) {
	list_type temp;

	while (l->next != NIL && ! list_elt_equal (e, l->next->elt))
		l = l->next;
	if (!list_elt_equal (e, l->next->elt)) {
		myerror ("list_delete: element not in list");
		return;
	}
	temp = l->next->next;
	list_free_cell (l->next);
	l->next = temp;
	return;
}

void list_delete_nth (list_type l, int n) {
	list_type temp;

	while (l->next != NIL && --n > 0)
		l = l->next;
	if (l->next == NIL) return;
	temp = l->next->next;
	list_free_cell (l->next);
	l->next = temp;
	return;
}

void list_change_nth_to (list_type l, int n, list_elt_type e) {
	while (l->next != NIL && n-- > 0)
		l = l->next;
	l->elt = e;
	return;
}

list_elt_type list_first (list_type l) {
	return l->next->elt;
}

list_elt_type list_nth (list_type l, int n) {
	while ((l->next) != NIL && n-- > 0)
		l = l->next;
	return l -> elt;
}

void list_free (list_type l) {
	list_type rest;

	if (l == NIL) return;
	while ((rest = (l -> next)) != NIL) {
		list_free_cell (l);
		l = rest;
	}
	list_free_cell (l);
	return;
}

boolean list_equal (list_type l1, list_type l2) {
	for (;;) {
		l1 = l1->next;
		l2 = l2->next;
		if (l1 == NIL && l2 == NIL) return TRUE;
		else if (l1 == NIL || l2 == NIL
			 || ! list_elt_equal (l1->elt,l2->elt)) return FALSE;
	}
}

list_type list_copy (list_type l1) {
	list_type l2, cell, l;

	l = l2 = list_make ();
	while ((l1 = l1 -> next) != NIL) {
		cell = list_get_cell ();
		cell->elt = l1->elt;
		l2->next = cell;
		l2 = cell;
	}
	l2->next = NIL;
	return l;
}

void list_clobber (list_type l1, list_type l2) {
	if (l1->next) list_free (l1->next);
	l1->next = l2->next;
	list_free_cell (l2);
	return;
}

list_type list_map (function_one_type f, list_type l1) {
	list_type l2;

	l2 = list_make ();
	while ((l1 = (l1->next)) != NIL)
		list_insert (l2, (*f)(l1->elt));
	return l2;
}

list_type list_mapll (function_two_type f, list_type l1, list_type l2) {
	list_type l3;

	l3 = list_make();
	while ((l1 = (l1->next)) != NIL && (l2 = (l2->next)) != NIL)
		list_insert (l3, (*f)(l1->elt, l2->elt));
	return l3;
}

list_type list_mapcomb (function_two_type f, list_type l1, list_type l2) {
	list_type l, l3;

	l3 = list_make();
	while ((l1 = (l1->next)) != NIL) {
		l = l2;
		while ((l = (l->next)) != NIL) {
			list_insert (l3, (*f)(l1->elt, l->elt));
		}
	}
	return l3;
}

list_type list_maplc (function_two_type f, list_type l1, list_elt_type e) {
	list_type l2;
	l2 = list_make();
	while ((l1 = (l1->next)) != NIL)
		list_insert (l2, (*f)(l1->elt, e));
	return l2;
}

list_type list_mapcl (function_two_type f, list_elt_type e, list_type l1) {
	list_type l2;
	l2 = list_make();
	while ((l1 = (l1 -> next)) != NIL)
		list_insert (l2, (*f)(e, l1->elt));
	return l2;
}

boolean list_every (predicate_type ok, list_type l) 
{
    while ((l = (l -> next)) != NIL)
	if (!(*ok)(l->elt)) return FALSE;
    return TRUE;
}

/* This combines two lists that are assumed to have positive integer elements
 * into a single list, where the elements of the second list are distinguished
 * by negating them.  Only use is to map two lists into one so that a game
 * (with lists of left and right followers) can be hashed using the hash
 * function which operates on only a one list argument.  * Cluge *
 */
list_type list_combine (list_type l1, list_type l2) {
	list_type l3;
	l3 = list_copy (l1);
	while ((l2 = (l2 -> next)) != NIL)
		list_insert (l3, -(l2 -> elt));
	return l3;

}

list_type list_union (list_type l1, list_type l2) {
	list_type l3;
	l3 = list_copy (l1);
	while ((l2 = (l2 -> next)) != NIL)
		if (!list_member (l3, l2->elt)) list_insert (l3, l2->elt);
	return l3;
}

list_type list_multi_union (list_type l1, list_type l2) {
	l1 = list_copy (l1);
	while ((l2 = (l2 -> next)) != NIL)
		list_insert (l1, l2->elt);
	return l1;
}

unsigned long list_stat(void) {
	list_type l;
	long unsigned n=0, t;

	if (list_free_list != NIL) n = 1;
	for (l = list_free_list; l->next != NIL; l = l->next) n++;
	t = list_length (list_0) + list_length (ll1) + list_length (ll2) + list_length (ll3) + 4;
	printf ("free list: %lu; fixed lists: %lu\n", n, t);
	printf ("total list cells malloced: %lu\n", n + list_count);
	return n + t;
}
