/* Copyright (C) David Wolfe, 1990.  All rights reserved. */

/* WHOLE_TREE 1
   EVALUATED_TREE 2 */

#include <stdlib.h>
#include <stdio.h>
#include "externs.h"
#include "rational.h"
#include "list.h"
#include "gameops.h"
#include "output.h"
#define E_LBRACE (ss (TeX?"\\{":"{"))
#define E_RBRACE (ss (TeX?"\\}":"}"))
#define E_BAR (ss (TeX?"|":"|"))

char			chill_string[MAXBUFF];
char			warm_string[MAXBUFF];
char			map_string[MAXBUFF];
int expression (char *, game_type, int, boolean, boolean);
game_type		last_left_zeros_remainder;
game_type		last_right_zeros_remainder;


/*  These functions determine if the game g should be treated as 0^n|h */
int count_left_zeros (game_type g) {
	last_left_zeros_remainder = g;
	if (   count_left_options (g) == 1
	    && count_right_options (g) == 1
	    && eq (nth_left_option (g,1), zero)
	    && !(is_tiny(g) || is_miny(g) || is_num_up_star (g) || (is_on (g) && (flags & ONS_FLAG))))
		return 1+count_left_zeros (nth_right_option (g,1));
	else return 0;
}
int count_right_zeros (game_type g) {
	last_right_zeros_remainder = g;
	if (   count_left_options (g) == 1
	    && count_right_options (g) == 1
	    && eq (nth_right_option (g,1), zero)
	    && !(is_tiny(g) || is_miny(g) || is_num_up_star (g) || (is_on (g) && (flags & ONS_FLAG))))
		return 1+count_right_zeros (nth_left_option (g,1));
	else return 0;
}

boolean is_print_simple (game_type g) {
	static char s[MAXBUFF];
	return !expression (s, g, MAXBUFF-1, FALSE, TRUE);
}

void game_print (game_type g) {
	static char s[MAXBUFF];
	printf ("%s", game_sprintn (s, g, MAXBUFF-1));
	return;
}

void game_printf (game_type g) {
	static char s[MAXBUFF];
	printf ("%s\n", game_sprintn (s, g, MAXBUFF-1));
	return;
}

char *game_sprint (char *s, game_type g) {
	return game_sprintn (s, g, MAXBUFF-1);
}

static char *strsub (s, old, New)
	char *s;
	char old, New;
{
	char *t;
	for (t=s; *t != '\0'; t++)
		if (*t == old) *t = New;
	return s;
}

static game_type warm(game_type g) {
	set_var (dummy_var = make_var (ARGUMENT_SYMBOL), g);
	strsub (warm_string, 'g', *ARGUMENT_SYMBOL);
	g = game_parse (warm_string);
	strsub (warm_string, *ARGUMENT_SYMBOL, 'g');
	return g;
}

static game_type chill (game_type g) {
	set_var (dummy_var = make_var (ARGUMENT_SYMBOL), g);
	strsub (chill_string, 'g', *ARGUMENT_SYMBOL);
	g = game_parse (chill_string);
	strsub (chill_string, *ARGUMENT_SYMBOL, 'g');
	return g;
}

static game_type map_game (game_type g) {
	set_var (dummy_var = make_var (ARGUMENT_SYMBOL), g);
	strsub (map_string, 'g', *ARGUMENT_SYMBOL);
	g = game_parse (map_string);
	strsub (map_string, *ARGUMENT_SYMBOL, 'g');
	return g;
}

#define appendg(g,t) expression (s + strlen (s), g, n - strlen (s), t, TRUE)
char *game_sprintn (char *s, game_type g, int n) {
	game_type chilled, warmed, frac, gnorm, small;
	Q_type k, k2;
	int t, norm;
	
	strcpy (s, "");
	if (g && (flags & EVALUATE_FLAG)) {
		k = mean (g);
		t=FALSE;
		if (Q_ge (k, 0)) {
			k2 = Q_plus (k, Q(1,2));
			norm = Q_p (k2) / Q_q(k2);
			if (Q_q (k2) == 1) norm--;
		}
		else {
			k2 = Q_minus (k, Q(1,2));
			norm = - ((- Q_p (k2)) / Q_q (k2));
			if (Q_q (k2) == 1) norm++;
		}
		chilled = chill (minus (g, g_int (norm)));
		warmed = warm (chilled);
		frac = (!Q_eq (left_stop (chilled), right_stop (chilled)))
		    ? zero : num (mean (chilled));
		small = minus (chilled, frac);
		if (   eq (g, plus (warmed, gnorm = g_int (norm)))
		    || eq (g, plus (warmed, gnorm = plus (gnorm, star(1))))) {
			t = flags & EXPAND_FLAG;
			flags = (flags | EXPAND_FLAG) ^ EXPAND_FLAG;	/* clear flag */
			if (eq (chilled, zero))
				appendg (gnorm, FALSE);
			else {
				if (!eq (gnorm, zero)) appendg (gnorm, FALSE);
				if (strlen (s) < n) strcat (s, flags & TEX_FLAG ? "\\int " : "%");
				if (!eq (frac, zero)) appendg (frac, FALSE);
				if (!eq (small, zero)) appendg (small, TRUE);
			}
			flags = flags | t;
		} else appendg (g, FALSE);
	} else
	  appendg (map_game (g), FALSE);
	if (strlen (s) == n) strcpy (s+n-3, "...");
	return s;
}

void list_print (list_type l) {
	static char s[MAXBUFF];
	printf ("%s", list_sprintn (s, l, MAXBUFF-1));
	return;
}

void list_printf (list_type l) {
	static char s[MAXBUFF];
	printf ("%s\n", list_sprintn (s, l, MAXBUFF-1));
	return;
}

void games_printf (list_type l) {
  printf ("( ");
  while (l->next != NIL) {
    l = l->next;
    game_printf (l->elt);
    printf (" ");
  }
  printf (")\n");
}

char *list_sprint (char *s, list_type l) {
	return list_sprintn (s, l, MAXBUFF-1);
}

char *list_sprintn (char *s, list_type l, int n) {
	static char t[20];

	strncpy (s, "( ", n - strlen (s));
	while (l->next != NIL) {
		l = l->next;
		sprintf (t,"%d ", l->elt);
		strncat (s, t, n - strlen (s));
	}
	strncat (s, ")", n - strlen (s));
	return s;
}

#define ss(x) (strncat (s, (x), n - strlen (s)))
#define sd(x) (sprintf (buff, flags & TEX_FLAG ? " %d" : "%d", (x)), ss(buff))
#define sg(x) (expression (buff, (x), MAXBUFF-1, FALSE, FALSE), ss (buff))
#define sgp(x) (expression (buff, (x), MAXBUFF-1, TRUE, FALSE), ss (buff))
#define sv(x) (var_sprintn (s, var_value (x), n - strlen (s)))
#define TeX (flags & TEX_FLAG)
int expression (char *s, game_type g, int n, boolean brace, boolean attop) {
	char buff[MAXBUFF], left[MAXBUFF], right[MAXBUFF];
	int t, bars, lbars=0, rbars=0, i, ls, rs, lz=0, rz=0;
	Q_type m;
	game_type h;

	strcpy (s, "");
	h = minus (g, num (m = left_stop (g)));
	ls = count_left_options (g);
	rs = count_right_options (g);
	if (   ((is_tiny (h) || is_miny (h)) && (flags & TINYS_FLAG))
	    || ((is_on (h)) && (flags & ONS_FLAG))
	    || (is_num_up_star (g)
		&& (Q_eq (num_value (g), Q_0)	|| (flags & INTEGERS_FLAG))
		&& (Q_is_int (num_value (g))	|| (flags & NUMBERS_FLAG))
		&& ((up_value (g) == 0)		|| (flags & UPS_FLAG))
		&& ((star_value (g) == 0)	|| (flags & STARS_FLAG))))
	    {
		if (!Q_eq (m, Q_0) || eq (g, zero)) {
			if (TeX && Q_lt (m, Q_0)) ss ("\\Neg"), m = Q_negative (m);
			if (Q_is_int (m)) sd (Q_p(m));
			else if (TeX) (ss ("{"), sd (Q_p(m)), ss ("\\over"), sd (Q_q(m)), ss ("}"));
			else (sd (Q_p(m)), ss ("/"), sd (Q_q(m)));
		}
		if ((t = up_value (g)) != 0)
			if (TeX) {
				if (t == 2) ss ("\\Upup");
				else if (t == -2) ss ("\\Downdown");
				else if (t == 1) ss ("\\Up");
				else if (t == -1) ss ("\\Down");
				else ss (t>0? "\\UpN{" : "\\DownN{"), (t>1? sd (t) : t<-1? sd(-t) :0), ss("}");
			} else ss (t>0? "^" : "v"), (t>1? sd(t) : t<-1? sd(-t) :0);
		if ((t = star_value (g)) > 0) (ss (TeX?"\\Star":"*"), t>1? sd (t) :0);
		if (is_tiny (h)) ss (TeX?"\\Tiny{":"+["), sg (tiny_value (h)), ss (TeX?"}":"]");
		if (is_miny (h)) ss (TeX?"\\Miny{":"-["), sg (miny_value (h)), ss (TeX?"}":"]");
/*		if (is_on (h)) sgp (on_value (h)), ss (TeX?"^{\\to ":"<("), sd (on_exponent (h)), ss (TeX?"}":")>");  (Old form) */
		if (is_on (h)) sgp (on_value (h)), ss (TeX?"^{\\to ":"<to "), sd (on_exponent (h)), ss (TeX?"}":">");
		return 0;
	} else if (   (flags & ZEROS_FLAG)
		   && rs == 1
		   && ls == 1
		   && ((lz = count_left_zeros (g)) > 1 || (rz = count_right_zeros (g)) > 1)) {
		if (brace) E_LBRACE;
		if (lz > 1) {
			ss (TeX ? " 0^{" : "0<");
			sd (lz);
			ss (TeX ? "} | " : ">|");
			sgp (last_left_zeros_remainder);
		} else {
			sgp (last_right_zeros_remainder);
			ss (TeX ? "| 0^{" : "|0<");
			sd (rz);
			ss (TeX ? "}" : ">");
		}
		if (brace) E_RBRACE;
		return 1;
	} else if (   Q_eq (m, right_stop(g))
		   && !Q_eq(m, Q_0)
		   && (Q_is_int (m) || (flags & NUMBERS_FLAG))) {
		sg (num (m));
		sgp (h);
		return 0;
	} else if ((flags & VARS_FLAG) && is_var (g)) {
		sv (g);
		if (attop && (flags & EXPAND_FLAG)) ss (" = ");
		else return 0;
	}
	/*********** At this point we're looking at a Gl|Gr or {Gl|Gr} **********/
	*left = *right = '\0';
	if (ls == 0 || rs == 0) brace = TRUE;
	for (i = 1; i <= ls; i++) {
		t = expression (buff, nth_left_option (g, i), MAXBUFF-2, ls > 1, FALSE);
		if (i < ls) strcat (buff, ",");
		lbars = (t > lbars) ? t : lbars;
		strncat (left, buff, MAXBUFF-1 - strlen (left));
	}
	for (i = 1; i <= rs; i++) {
		t = expression (buff, nth_right_option (g, i), MAXBUFF-2, rs > 1, FALSE);
		if (i < rs) strcat (buff, ",");
		rbars = (t > rbars) ? t : rbars;
		strncat (right, buff, MAXBUFF-1 - strlen (right));
	}
	lbars++;
	rbars++;
	bars = (lbars > MAXBARS ? rbars :
		rbars > MAXBARS ? lbars :
		lbars > rbars ? lbars : rbars);
	if (brace) E_LBRACE;
	if (lbars > MAXBARS) E_LBRACE;
	ss (left);
	if (lbars > MAXBARS) E_RBRACE;
	if (bars > MAXBARS) E_BAR;
	else for (i = 1; i <= bars; i++) E_BAR;
	if (rbars > MAXBARS) E_LBRACE;
	ss (right);
	if (rbars > MAXBARS) E_RBRACE;
	if (brace) E_RBRACE;
	return (brace ? 0 : bars > MAXBARS ? 1 : bars);
}
