/* A sample program to evaluate the game of Wyt Queens. */

/* Define a key for the hash table. Keys below 1000 are reserved. */
#define WYT_QUEENS_KEY 1001

#include "games.c"	/* Includes all the needed .h files */

game_type play_wyt_queens (int x, int y) {
	list_type position_as_list;
	list_type left, right;
	game_type g;
	int n, min;

	/* Encode the position as a list of integers, in order use the */
	/* hash table to store computed positions.  If the position */
	/* has aready been evaluated, return the precomputed value. */
	position_as_list = list_make();
	list_prepend (position_as_list, x);
	list_prepend (position_as_list, y);
	if (hash_test (WYT_QUEENS_KEY, position_as_list)) {
		list_free (position_as_list);	/* free the space used by the */
						/* list */
		return hash_get_last();
	}
	
	/* Position wasn't already computed, so evaluate the position recursively. */
	left = list_make();
	for (n=0; n<x; n++) list_insert (left, play_wyt_queens (n, y)); /* Horizontal moves */
	for (n=0; n<y; n++) list_insert (left, play_wyt_queens (x, n)); /* Vertical */
	min = ( x<y ? x : y);
	for (n=1; n<=min; n++) list_insert (left, play_wyt_queens (x-n,y-n)); /* Diagonal */
	right = list_copy (left);	/* Right's options are the same as Left's. */
	g = make (left, right);		/* Construct the canonical form of the */
					/* game.  Lists left and right are destroyed. */

	/* Store the value of the position in hash table.  The position_as_list */
	/* is destroyed and freed by hash_put() by, so no need to list_free() it. */
	hash_put (WYT_QUEENS_KEY, position_as_list, g);
	return g;
}

#define print_dashes(n) for (i=1; i<=n; i++) printf ("-")

void main() {
	char s[10];

	int size = 20;
	int i,j;
	init();			/* initialize the toolkit! */

	game_sprintn (s, play_wyt_queens(2,6), 9);
	printf ("A queen at location (2,6) has value %s\n", s);

	printf ("\n%4s | ", "");
	for (j=1;j<=size;j++) printf ("%4d", j);
	printf ("\n");
	print_dashes(4);
	printf ("---");
	for (j=1;j<=size;j++) print_dashes(4);
	for (i=1; i<=size; i++) {
	    printf ("\n%4d | ", i);
	    for (j=1;j<=size; j++) {
		printf ("%4s", game_sprintn (s, play_wyt_queens(i,j), 9));
	    }
	}
	printf ("\n");
}
